#
# Perl Module for Aceclient
#
# Cared for by Andy Williams <aw4@sanger.ac.uk>
#
#Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE
#

package AceTagClient;

use vars qw($AUTOLOAD @ISA @EXPORT_OK);
use Exporter;
use Carp;
#use strict;

use AceTagSet;
use Aceclient;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

#
# @ISA has our inheritance.
#

@ISA = ( 'Exporter' );

my %fields = (
	      server  => undef,
	      port => undef,
	      );

sub new {
    
    my $self = {
	'_permitted' => \%fields,
	%fields, 
    };
    bless $self;
    return $self;

}

sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) || carp "Ok, $self is not an object of mine!";
    my $name = $AUTOLOAD;

    # don't propagate DESTROY messages...

    $name =~ /::DESTROY/ && return;

    $name =~ s/.*://; #get only the bit we want
    unless (exists $self->{'_permitted'}->{$name} ) {
	carp "In type $type, can't access $name - probably passed a wrong variable into Sequence  ";
    }
    if (@_) {
	return $self->{$name} = shift;
    } else {
	return $self->{$name}
    }
}


sub connect {
    my $self = shift;
    my $server = shift;
    my $port = shift;
    my $time_out = shift;
    my $ace;
    my $query = "quit";

    if ($ace = openServer($server, $port, $time_out)) {
	# If connects ok
	$self->server($server);
	$self->port($port);
	# Return handle
	return $ace;
    } else {
	return 1;
    }
   
    
}

sub query {
    my $self = shift;
    my $ace = shift;
    my $query = shift;
    my $chunk_size = shift;
    my (@result, @query_lines, @lines);
    my $j;
    # Query must be seperated by &
    
    my $cnt = $query =~ s/\&/\&/g;
    @query_lines = split(/\&/, $query);

    if (! defined $chunk_size) {
	$chunk_size = 0;
    }

    for ($j = 0; $j<=$cnt; $j++) {	
	my ($status, $answer) = askServer($ace, $query_lines[$j], $chunk_size);
	push @result, $answer;
    }
    
    #Testing...

    #print "RESULT = @result\n\n\n";

    my $result = new AceTagSet;

    foreach (@result) {
	@lines = split(/\n/);
	foreach (@lines) {
	    if (! /(\/\/)|(^acedb)/) {
		($a, $b) = split (/\s+/, $_, 2);
		$a =~ s/\s+//g;
		$b =~ s/\"//g;
		$b =~ s/\s\:\s/\s/g;
		chomp($b); 
		$result->AddTagValuePair($a,$b);
		
	    } else {
		$result->AddTagValuePair("LOG",$_);
		
	    }
	}
	
    }

    return $result; # give back a AceTagSet
}

sub disconnect {

    my $self = shift;
    my $ace = shift;
    
    closeServer($ace);

    return 0;

}



1;  # says use was ok

__END__

=head1 NAME

AceTagClient

=head1 DESCRIPTION

B<AceTagClient.pm> Allows access to an ACEDB database. Allows
quering of any ACEDB datbase. 

It uses Steffan Berg's Aceclient module and puts results into a hash
array, which is an AceTagSet (see further information below). The
AceTagSet object is a reasonably transparent representation of
what you can get out of an acedb query.



=head1 AUTHOR

B<Andy Williams> Email aw4@sanger.ac.uk

=over

=head2 Example1

A simple example to get us started.

		#!/usr/local/bin/perl

		use AceTagClient;

		# Create a new connection to ACEDB

		$test = new AceTagClient;

		$ace_handle = $test->connect("humsrv1", "210000", "1200");

                # Check that the connection succeeded.
                
                if ($ace_handle == 1) {
                     print "Connection refused : $!\n";
                     exit(1);
                }

		# Show everything about a clone.

		$command = "find genome_sequence cN117F11 & show -a & quit";

		$results = $test->query($ace_handle, $command);

                # Disconnect from server.

                $test->disconnect($ace_handle);

                foreach $tag ($result->EachTag()) {
    
		    $name = $tag->name();
		    print "$name\n";
		    foreach $value ($tag->EachValue()) {
			print "\t$value\n";
		    }
		}

                # Find a specific tag.
 
                $tag = $results->GetTag("DNA_Homol");
                foreach $value ($tag->EachValue()) {
                    print "\t$value\n";
                }

=head2 Making a new AceTagClient

=over

=item C<new>

C<AceTagClient::new>

Creates connection.

=item C<connect>

B<connect($server, $port, $time_out)>.Connects to a specified server and 

port number. Returns the handle to aceclient. Returns 1 if connection fails.

=head1 Functions

=item C<query>

B<query($handle, $query, $chunk_size)>. Queries the current ACEDB and returns
results, without comments, in a hash array - hashed on the tags. 
$query should in the form B<find sequence & show -a> ie seperated
by '&'.
A $chunk_size of 5-10(kb) is recommended by Stephan Berg but I found
that it works just as well with size 0 (in which case you can ommit the
chunk size altogether).

=back

=item C<disconnect>

B<disconnect($handle)>. Closes the connection to ACEDB.

=back

=head1 Further Information

=for html
<a href="Aceclient.html">Aceclient</a>

=for html
<a href="AceTagSet.html">AceTagSet</a>

=for html
<a href="AceTag.html">AceTag</a>









