package Bio::GSC::Feature;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Feature

=head1 DESCRIPTION

Bio::GSC::Feature is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Feature is the base class for sequence features. There are
two types of sequence features, simple and complex. Simple features
minimally contain Begin and End attributes. Complex features contain a
group of features, called SubFeatures. A simple feature might be the
location of an Alu repeat while a complex feature might be a gene
prediction containing exon SubFeatures.

=head1 DEPENDENCIES

Bio::GSC::Utilities

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use Bio::GSC::Constants;
use Bio::GSC::ScoringMatrix;
use Bio::GSC::Utilities qw(getFasta fcmp);
use Bio::GSC::NT qw(antiDNA);

#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------
my $FIELD  = '\S+';
my $NOP    = '<|>|<=|>=|==';   # number operators
my $TOP    = 'eq|ne|gt|lt';    # text operators
my $POP    = '=~|!~';          # pattern operator
my $TEXT   = '\'[\w\s]+\'';
my $MATCH  = '/[\w\s]+/';
my $NUMBER = '[\de\.\-\+]+';

#----------------------------------------------------------------------
# new
#----------------------------------------------------------------------

=head2 new()

Description: Instantiates a feature. Note: not usually called on its
own, but by Bio::GSC::Tool::TOOL_TYPE::parse(), which creates features.
See Bio::GSC::Tool for more information on Tools.

 Usage: $feature = new Bio::GSC::Feature();

=cut

sub new {
	my $f = {};
	return bless $f;
}

#----------------------------------------------------------------------
# clone
#----------------------------------------------------------------------

=head2 clone()

Description: The Feature makes a copy of itself.

 Usage: $duplicate = $feature->clone();

=cut

sub clone {
	my ($this) = @_;
	my $dup = new;
	foreach my $field (keys %$this) {
		$dup->{$field} = $this->{$field} unless ref($this->{$field});
	}
	return $dup unless $this->isComplex();

	my $subfeature = $this->getSubFeature();
	foreach my $sf (@$subfeature) {
		my $sfdup = $sf->clone();
		$dup->addSubFeature($sfdup);
	}
	return $dup;
}

#----------------------------------------------------------------------
# Attribute Accessor/Assignment Methods
#----------------------------------------------------------------------

#----------------------------------------------------------------------
# Name
#----------------------------------------------------------------------

=head2 Name Attribute

Description: The Name attribute is optional. Some features, like a
region of low complexity has no name, while a region containing a
repetitive element usually does (eg. Alu repeat). Name will default to
SbjctDef if no Name is given found.

 Usage: $feature->getName();
        $feature->setName($name);

=cut

sub getName {
	 my ($this) = @_;
	 if (defined $this->{'Name'}) {return $this->{'Name'}}
	 elsif (defined $this->getSbjctDef) {return $this->getSbjctDef}
	 else {return ""}
}

sub setName {
	 my ($this, $value) = @_;
	 $this->{'Name'} = $value;
}

#----------------------------------------------------------------------
# Begin and End
#----------------------------------------------------------------------

=head2 Begin and End Attributes

Description: All features have Begin and End coordinates. If End is
greater, the feature is on the complementary strand. Begin and End
are computed from SubFeatures if a feature is complex (contains
SubFeatures).

 Usage: $feature->getBegin();
        $feature->setBegin($begin);
        $feature->getEnd();
        $feature->setEnd($end);

=cut

sub getBegin {
	my ($this) = @_;
	if ($this->isComplex()) {
		if ($this->getStrand() > 0)
			{return $this->minAtt('Begin')}
		elsif ($this->getStrand() < 0)
			{return $this->maxAtt('Begin')}
	}
	else {return $this->{'Begin'}}
}

sub setBegin {
	my ($this, $value) = @_;
	$this->{'Begin'} = $value;
}

sub getEnd {
	my ($this) = @_;
	if ($this->isComplex()) {
		if ($this->getStrand() > 0)
			{return $this->maxAtt('End')}
		elsif ($this->getStrand() < 0)
			{return $this->minAtt('End')}
	}
	else {return $this->{'End'}}
}

sub setEnd {
	my ($this, $value) = @_;
	$this->{'End'} = $value;
}

#----------------------------------------------------------------------
# Score, P, Percent
#----------------------------------------------------------------------

=head2 Score, P, and Percent Attributes

Description: Predictions and alignment algorithms commonly use certain
attributes such as Score, P, and Percent. Score is any number, P is
from 0 to 1.0, and Percent ranges from 0 to 100.

 Usage: $feature->getScore();
        $feature->setScore($score)
        $feature->getP();
        $feature->setP($probability);
        $feature->getPercent();
        $feature->setPercent($percent);

=cut

sub getScore {
	my ($this) = @_;
	if ($this->isComplex()) {return $this->sumAtt('Score')}
	else {return $this->{'Score'}}
}

sub setScore {
	my ($this, $value) = @_;
	$this->{'Score'} = $value;
}

sub getP {
	my ($this) = @_;
	if ($this->isComplex()) {return $this->minAtt('P')}
	else {return $this->{'P'}}
}

sub setP {
	my ($this, $value) = @_;
	$this->{'P'} = $value;
}

sub getPercent {
	my ($this) = @_;
	if ($this->isComplex()) {return $this->maxAtt('Percent')}
	{return $this->{'Percent'}}
}

sub setPercent {
	my ($this, $value) = @_;
	$this->{'Percent'} = $value;
}

#----------------------------------------------------------------------
# Length, Strand, and Span
#----------------------------------------------------------------------

=head2 Length, Strand, and Span Attributes

Description: These are accessor only methods. For simple features,
both Length and Span are the same. Strand is +1 if  Begin < End, and
-1 if Begin > End.
For complex features, Begin and End are computed using the SubFeatures,
and Span is the distance  between these. Length is the sum of the
individual feature lengths.

 Usage: $feature->getLength();
        $feature->getStrand();
        $feature->getSpan();

=cut

sub getLength {
	my ($this) = @_;
	if ($this->isComplex()) {return $this->sumAtt('Length')}
	else {return abs($this->getBegin() - $this->getEnd()) + 1}
}

sub getStrand {
	my ($this) = @_;
	if ($this->isComplex()) {
		#----- compute from strand of first subfeature
		my @sf = @{$this->getSubFeature()};
		return $sf[0]->getStrand();
	}
	return $this->getBegin() > $this->getEnd() ? -1 : 1;
}

sub getSpan {
	my ($this) = @_;
	return abs($this->getBegin - $this->getEnd) + 1;
}


#----------------------------------------------------------------------
# isComplex
#----------------------------------------------------------------------

=head2 isComplex()

Description: determines if a feature is simple or complex (containing
SubFeatures).

 Usage: $feature->isComplex();

=cut

sub isComplex {
	my ($this) = @_;
	if (defined $this->{'SubFeature'}) {return TRUE}
	else {return FALSE}
}

sub getisComplex {isComplex(@_)} # just so it works in queryFeature

#----------------------------------------------------------------------
# SubFeature
#----------------------------------------------------------------------

=head2 SubFeature Attribute

Description: SubFeature contains more features. For example, a gene is
a feature that contains exon subfeatures. Similarly, a pairwise
local alignment feature may contain several alignments.

 Usage: $feature->getSubFeature();
        $feature->setSubFeature(\@list_of_features);
        $feature->addSubFeature($feature);

=cut

sub getSubFeature {
	my ($this) = @_;
	return $this->{'SubFeature'};
	# should return by sorting? YES!!!! by strand and begin
}

sub setSubFeature {
	my ($this, $value) = @_;
	$this->{'SubFeature'} = $value; # better be a list ref!
}

sub addSubFeature {
	my ($this, $value) = @_;
	push(@{$this->{'SubFeature'}}, $value);
}

#----------------------------------------------------------------------
# minAtt, maxAtt, sumAtt, and productAtt
#----------------------------------------------------------------------

=head2 minAtt, maxAtt, sumAtt, and productAtt Attributes

Description: These accessor methods compute the minimum, maximum
sum, or product of a complex feature (a feature containing SubFeatures).

 Usage: $minimum = $feature->minAtt('Length');
        $maximum = $feature->maxAtt('Begin');
        $sum = $feature->sumAtt('Score');
        $product = $feature->productAtt('P');

=cut

sub minAtt {
	my ($this, $att) = @_;
	my @sf = @{$this->getSubFeature()};
	my $min = 1.0e100;
	foreach my $sub (@sf) {
		my $val = eval "\$sub->get$att()";
		next unless defined $val;
		$min = $val if $val < $min;
	}
	return $min;
}

sub maxAtt {
	my ($this, $att) = @_;
	my @sf = @{$this->getSubFeature()};
	my $max = -1.0e100;
	foreach my $sub (@sf) {
		my $val = eval "\$sub->get$att()";
		next unless defined $val;
		$max = $val if $val > $max;
	}
	return $max;
}

sub sumAtt {
	my ($this, $att) = @_;
	my @sf = @{$this->getSubFeature()};
	my $sum = 0;
	foreach my $sub (@sf) {
		my $val = eval "\$sub->get$att()";
		next unless defined $val;
		$sum += $val;
	}
	return $sum;
}

sub productAtt {
	my ($this, $att) = @_;
	my @sf = @{$this->getSubFeature()};
	my $sum = 0;
	foreach my $sub (@sf) {
		my $val = eval "\$sub->get$att()";
		next unless defined $val;
		$sum += $val;
	}
	return $sum;
}

#----- queryFeature Hacks -----#
sub getminATT     {minATT(@_)}
sub getmaxATT     {maxATT(@_)}
sub getsumATT     {sumATT(@_)}
sub getproductATT {productATT(@_)}

#----------------------------------------------------------------------
# SbjctID, SbjctDef, SbjctBegin, SbjctEnd
#----------------------------------------------------------------------

=head2 SbjctID, SbjctDef, SbjctBegin, SbjctEnd, and SbjctStrand

Description: These attributes are used for pairwise alignments (eg.
BLAST).

 Usage: $feature->getSbjctID();
        $feature->setSbjctID($gi);
        $feature->getSbjctDef();
        $feature->setSbjctDef($name);
        $feature->getSbjctBegin();
        $feature->setSbjctBegin($begin);
        $feature->getSbjctEnd();
        $feature->setSbjctEnd($end);
		$feature->getSbjctStrand();

=cut

sub getSbjctID {
	my ($this) = @_;
	return defined $this->{'SbjctID'} ? $this->{'SbjctID'} : "";
}
sub setSbjctID {
	my ($this, $value) = @_;
	$this->{'SbjctID'} = $value;
}

sub getSbjctDef {
	my ($this) = @_;
	return $this->{'SbjctName'};
}
sub setSbjctDef {
	my ($this, $value) = @_;
	$this->{'SbjctName'} = $value;
}

sub getSbjctBegin {
	my ($this) = @_;
	return $this->{'SbjctBegin'};
}
sub setSbjctBegin {
	my ($this, $value) = @_;
	$this->{'SbjctBegin'} = $value;
}

sub getSbjctEnd {
	my ($this) = @_;
	return $this->{'SbjctEnd'};
}
sub setSbjctEnd {
	my ($this, $value) = @_;
	$this->{'SbjctEnd'} = $value;
}

sub getSbjctStrand {
	my ($this) = @_;
	if    ($this->getSbjctBegin < $this->getSbjctEnd)   {return  1}
	elsif ($this->getSbjctEnd < $this->getSbjctBegin)   {return -1}
	else  {warn "ERROR: SbjctBegin == SbjctEnd\n"; return  0}
}
sub getSbjctLength {
	my ($this) = @_;
	if    ($this->getSbjctBegin < $this->getSbjctEnd) {
		return $this->getSbjctEnd - $this->getSbjctBegin + 1;
	}
	elsif ($this->getSbjctEnd < $this->getSbjctBegin) {
		return $this->getSbjctBegin - $this->getSbjctEnd + 1;
	}
	else  {warn "ERROR: SbjctBegin == SbjctEnd\n"; return  0}
}


#----------------------------------------------------------------------
# QueryGap, SbjctGap
#----------------------------------------------------------------------

=head2 QueryGap and SbjctGap Attributes

Description: These attributes are used for pairwise alignments (eg.
BLAST), and indicate the position and length of sequence gaps. Gaps are
stored as a HASH where the key is the coordinate and the value is the
number of gaps. The coordinate is relative to the alignment and not the
actual coordinate on either sequence.

 Usage: $q_gap = $feature->getQueryGap();
        $feature->setQueryGap(\%position_and_length);
        $s_gap = $feature->getSbjctGap();
        $feature->setSbjctGap(\%position_and_length);

=cut

sub getQueryGap {
	my ($this) = @_;
	return $this->{'QueryGap'};
}
sub setQueryGap {
	my ($this, $hash_ref) = @_;
	$this->{'QueryGap'} = $hash_ref;
}

sub getSbjctGap {
	my ($this) = @_;
	return $this->{'SbjctGap'};
}
sub setSbjctGap {
	my ($this, $hash_ref) = @_;
	$this->{'SbjctGap'} = $hash_ref;
}

#----------------------------------------------------------------------
# QuerySeq, SbjctSeq, Alignment, Identity, and Similarity
#----------------------------------------------------------------------

=head2 QuerySeq, SbjctSeq, Alignment, Identity, and Similarity

Description: These optional attributes are available for pairwise
alignments (eg. BLAST). QuerySeq, SbjctSeq, and Alignment hold the
actual alignments (gaps included). Any IUPAC codes are legal, as is a -
to indicate a sequence gap. In an alignment, + is also a legal value,
indicating a conservative change. Identitiy and Similarity methods
compute percentages based on the alignment sequence (which is based on
the scoring matrix used in the original program). In order to use these
pairwise-specific attributes, you must make a direct call to the class
method ParseSbjctSequences() of the particular parser. The default
behavior is not to parse SbjctSequences because they can be memory
intensive.

 Usage: use Bio::GSC;
		Bio:GSC::Tool::Blast::ParseSbjctSequences(TRUE);
        $feature->getQuerySeq();
        $feature->setQuerySeq($sequence);
        $feature->getSbjctSeq();
        $feature->setSbjctSeq($sequence);
        $feature->getAlignment();
        $feature->setAlignment($alignment);
        $feature->getIdentity();
        $feature->getSimilarity();

=cut

sub getQuerySeq {
	my ($this) = @_;
	return $this->{'QuerySeq'};
}
sub setQuerySeq {
	my ($this, $value) = @_;
	$this->{'QuerySeq'} = $value;
}

sub getSbjctSeq {
	my ($this) = @_;
	return $this->{'SbjctSeq'};
}
sub setSbjctSeq {
	my ($this, $value) = @_;
	$this->{'SbjctSeq'} = $value;
}

sub getAlignment {
	my ($this) = @_;
	return $this->{'Alignment'};
}
sub setAlignment {
	my ($this, $value) = @_;
	$this->{'Alignment'} = $value;
}

sub getIdentity {
	my ($this) = @_;
	my $align = $this->getAlignment();
	$align =~ tr/+/ /;
	my $neg = $align =~ tr/ / /;
	return 100 * (1 - int ($neg / length($align)));
}
sub getSimilarity {
	my ($this) = @_;
	my $align = $this->getAlignment();
	my $neg = $align =~ tr/ / /;
	return 100 * (1 - int ($neg / length($align)));
}

#----------------------------------------------------------------------
# queryFeature
#----------------------------------------------------------------------

=head2 queryFeature()

Description: allows one to query a feature. For example, one can ask
if a feature has a Score > X and a Length > Y. The syntax for such
statements follows standard perl syntax ("and" "or" "not" can be used
for compound statements and parentheses can be used to force
precedence). The attribute is used as a variable name. The return value
is either TRUE or FALSE (1 or 0). See the examples below. Note that you
must use single quotes for text comparison operators, and you must use
slashes for the pattern match operator. Any of the "getAttribute()"
methods will work in addition to isComplex() and the minAtt(), maxAtt(),
sumAtt(), and productAtt() methods.

 Usage: $feature->queryFeature($query_statement);
 Examples:
   "Begin > 100"
   "Name ne 'Alu' and Length > 200"
   "SbjctName =~ /sapiens/ and Strand > 1"
   "(Begin > 100 or End > 100) and Length > 50" 

=cut

sub queryFeature {
	my ($this, $filter) = @_;
	return 1 unless defined $filter or $filter;

	#----- parse filtering into eval-able code (see class variables)
	$filter =~ s/($FIELD) ($NOP) ($NUMBER)/\$this->get$1() $2 $3/g;
	$filter =~ s/($FIELD) ($TOP) ($TEXT)/\$this->get$1() $2 $3/g;
	$filter =~ s/($FIELD) ($POP) ($MATCH)/\$this->get$1() $2 $3/g;

	my $result = eval $filter;
	# do something with undefined values?
	return $result;
}

#----------------------------------------------------------------------
# Analysis Functions
#----------------------------------------------------------------------
my $Genomic = "";

=head2 SetGenomic()

Description: Sets the class variable used by the cmpByGenomic()
method. Must be reset between different analysis "sessions".

 Usage: Bio::GSC::Feature::SetGenomic($some_sequence);

=cut

sub SetGenomic {
	($Genomic) = @_;
}

sub _extractGenomic {
	my ($p, $q) = @_;
	#----- make sure the extracted part doesn't run over the ends
	$p = $p < 0 ? 0 : $p;
	$q = $q < 0 ? 0 : $q;
	$p = $p >= length($p) ? length($p) - 1 : $p;
	$q = $q >= length($q) ? length($q) - 1 : $q;
	#----- return REF to sequence or antiparallel of sequence
	my $seq;
	if ($p < $q)
		{$seq = substr($Genomic, $p, $q - $p)}
	else {
		$seq = substr($Genomic, $q, $p - $q);
		$seq = antiDNA($seq);
	}
	return \$seq;
}


#----------------------------------------------------------------------
# cmpByCoor
#----------------------------------------------------------------------

=head2 cmpByCoor()

Description: Compares a pair of features using their Begin and End
Attributes and a boundary. CmpByCoor() returns various values
depending upon how the first feature compares to the second feature.
See below for return values.

 Usage: $feature1->cmpByCoor($feature2, $boundary);
 Returns:
          gscIdentical - both coordinates are within fuzzy
          gscSuper     - longer on both ends
          gscLong5     - longer on 5' end only
          gscLong3     - longer on 3' end only
          gscShort5    - shorter on 5' end only
          gscShort3    - shorter on 3' end only
          gscSubset    - shorter on both ends
          gscSkew5     - longer on 5' end and shorter on 3' end
          gscSkew3     - longer on 3' end and shorter on 5' end
          0            - neither end is within fuzzy boundary or
                         features are on different strands

=cut

sub cmpByCoor {
	my ($f1, $f2, $fuzzy) = @_;
	my ($begin, $end, $result);

	# strand sensitive
	return 0 if $f1->getStrand() ne $f2->getStrand();

	# compare the coordinates
	$begin = fcmp($f1->getBegin(), $f2->getBegin(), $fuzzy);
	$end   = fcmp($f1->getEnd()  , $f2->getEnd()  , $fuzzy);

	# produce the results
	if ($begin ==  1 and $end ==  1)
		{return ($f1->getStrand() > 1) ?    gscSkew3 : gscSkew5}
	elsif ($begin ==  1 and $end ==  0)
		{return ($f1->getStrand() > 1) ?   gscShort5 : gscLong5}
	elsif ($begin ==  1 and $end == -1)
		{return ($f1->getStrand() > 1) ?   gscSubset : gscSuperset}
	elsif ($begin ==  0 and $end ==  1)
		{return ($f1->getStrand() > 1) ?    gscLong3 : gscShort3}
	elsif ($begin ==  0 and $end ==  0)
		{return                                gscIdentical}
	elsif ($begin ==  0 and $end == -1)
		{return ($f1->getStrand() > 1) ?   gscShort3 : gscLong3}
	elsif ($begin == -1 and $end ==  1)
		{return ($f1->getStrand() > 1) ? gscSuperset : gscSubset}
	elsif ($begin == -1 and $end ==  0)
		{return ($f1->getStrand() > 1) ?    gscLong5 : gscShort5}
	elsif ($begin == -1 and $end == -1)
		{return ($f1->getStrand() > 1) ?    gscSkew5 : gscSkew3}
}

#----------------------------------------------------------------------
# cmpByGenomic
#----------------------------------------------------------------------

=head2 cmpByGenomic()

Description: Compares a pair of sequence features using genomic
sequence for the comparison. The comparison acutally takes place at the
level of the genomic sequence. This means that for pairwise alignments,
the sequence of the subject isn't used. cmpByGenomic asks "does this
sequence contain that sequence or vice-versa?". Therefore it has fewer
return values than cmpByCoor(). Like other feature comparison
operators, cmpByGenomic requires two arguments: a feature object
and a boundary. Here, the boundary is the amount of extra sequence
extracted from the genomic. In addition, the genomic sequence must be
stored in this package using Bio::GSC::Feature::SetGenomic (see above).

 Usage: $feature1->cmpByGenomic($feature2, $boundary);
 Returns:
          gscIdentical - both are the same...relative to genomic
          gscSuperset  - first contains second
          gscSubset    - second contains first
          0            - features are on different strands or do not
                         "fit" each other (eg. skewed)

=cut

sub cmpByGenomic {
	my ($f1, $f2, $fuzzy) = @_;
	my ($sub1, $sub2, $result);
	my ($strand1, $strand2);

	# strand sensitive
	return 0 if $f1->getStrand() ne $f2->getStrand();

	# compare the sequences
	$sub1 = getGenomic($f1->getBegin(), $f1->getEnd());
	$sub2 = getGenomic($f2->getBegin(), $f2->getEnd());

	# produce the results
	if    ($$sub1 eq $$sub2)           {return gscIdentical}
	elsif (index($$sub1, $$sub2) > -1) {return gscSuperset}
	elsif (index($$sub2, $$sub1) > -1) {return gscSubset}
	else                               {return 0}
}

#----------------------------------------------------------------------
# cmpByScore - not finished yet
#----------------------------------------------------------------------

=head2 cmpByScore()

Description: compares a pair of sequence features by computing an
alignment score. You must supply two arguments, a feature object and a
threshold cutoff score for the comparison to return true. Note that the
default scoring matrix is +5 -4.  To change the matrix, you must
make a direct call to Bio::GSC::ScoringMatrix::SetMatrix(). A false
value returns FALSE while a true value returns the score of the best
alignment. Be forwarned that this function is not very fast as compared
to BLAST. For more information see Bio::GSC::Alignment. and 
Bio::GSC::ScoringMatrix. Note that like all other feature comparisons,
sequences on different strands can never be "the same".

 Usage: $feature1->cmpByScore($feature2, $cutoff);

=cut

sub cmpByScore {
	my ($f1, $f2, $cutoff);
	my ($seq1, $seq2) = ($f1->getSeq(), $f2->getSeq());
	my $score = Bio::GSC::ScorePair($seq1, $seq2);
	return $score > $cutoff ? $score : 0;
}

#----------------------------------------------------------------------
# cmpFeature
#----------------------------------------------------------------------

=head2 cmpFeature()

Description: Compares complex features to one another. You must choose
the comparison operator (gscCoordinates, gscGenomic, or gscScore). The
comparison operator determines whether cmpFeature uses cmpByCoor,
cmpByGenomic, or cmpByScore.
The return value of cmpFeatures() is the same as those of the simple
feature comparison operators (above) if the Features being
compared are simple. Complex features have an additional return value,
gscOverlap, and when using the gscScore operator, the return value is
the sum of the scores.

 Usage: $feature1->CmpComplex($feature2, $mode, $operator, $cutoff);
 Returns:
       gscIdentical - both features have identical SubFeatures
       gscOverlap   - the features have non-contiguous features in
                      common (eg. some mismatching interior "exons")
       gscsuper     - feature1 contains all of feature2
       gscSubset    - feature1 is completely within feature2
       gscLong5     - feature1 contains additional SubFeatures, and they
                      are located at the 5' end
       gscLong3     - additional SubFeatures on the 3' end
       gscShort5    - fewer SubFeatures on the 5' end
       gscShort3    - fewer SubFeatures on the 3' end
       gscSkew5     - more features on the 5' end and less on the 3' end
       gscSkew3     - fewer features on the 3' end an more on the 5' end
       0            - the features have nothing in common

=cut

sub cmpFeature {
	my ($f1, $f2, $mode, $op, $bound) = @_;

	#----- if both features are simple life is easy
	unless ($f1->isComplex() or $f2->isComplex()) {
		return $f1->cmpByCoor($f2, $bound) if $op == gscCoordinates;
		return $f1->cmpByGenomic($f2, $bound) if $op == gscGenomic;
		return $f1->cmpByScore($f2, $bound) if $op == gscScore;
	}

	#----- check if they are on the same strand
	return 0 unless $f1->getStrand() == $f2->getStrand();

	#----- complex features need to be expanded (only two levels now)
	my (@g1, @g2); # g is for gene - a simple way to think about it
	if ($f1->isComplex()) {
		my $sf = $f1->getSubFeature();
		@g1 = @$sf;
	}
	else {push @g1, $f1};

	if ($f2->isComplex()) {
		my $sf = $f2->getSubFeature();
		@g2 = @$sf;
	}
	else {push @g2, $f2};

	#------------------------------------------------------------------
	# compare e1 to the rest of gene2 until something matches
	# and then anchor the comparisons there. If nothing matches,
	# get another e1 and try again. All interior alignments must
	# be identical but not so for the ends if gscIncompleteFeatures
	#------------------------------------------------------------------

	#------------------------------------------------------------------
	# Some values used for determining the structure of the
	# aligned genes
	#------------------------------------------------------------------
	my $match  = FALSE;
	my $hole   = FALSE;
	my $long5  = FALSE;
	my $long3  = FALSE;
	my $short5 = FALSE;
	my $short3 = FALSE;

	#------------------------------------------------------------------
	# Find the first matching exons. 
	#------------------------------------------------------------------
	my ($s1, $s2); # the starting offsets of the first matching exons
	FIND_EXON: for(my $i=0;$i<@g1;$i++) {
		$s2 = _exonInGene($mode, $op, $bound, $g1[$i], @g2);
		if (defined $s2) {
			$s1 = $i;
			$match = TRUE;
			last FIND_EXON;
		}
	}
	return 0 unless $match;

	#------------------------------------------------------------------
	# Find out who was longer at the 5' end.
	#------------------------------------------------------------------
	if    ($s1 > $s2) {$short5 = TRUE}
	elsif ($s2 > $s1) {$long5  = TRUE}

	#------------------------------------------------------------------
	# Chop the genes off so they start at the same exon.
	#------------------------------------------------------------------
	splice @g1, 0, $s1;
	splice @g2, 0, $s2;

	#------------------------------------------------------------------
	# Get an exon from each gene and compare them. If they match, then
	# crop the genes at their 5' ends. Keep repeating until one runs
	# out of exons (or both at the same time).
	#------------------------------------------------------------------
	ALIGN: {
		my $e1 = shift @g1;
		if (not defined $e1) {
			$short3 = TRUE if @g2; # gene1 is completely chopped. If
			last ALIGN;            # gene2 still exists, it is longer.
		}

		my $next_exon = _exonInGene($mode, $op, $bound, $e1, @g2);

		if (not defined $next_exon) {
			$long3 = TRUE;    # gene2 is completely chopped, but gene1
			last ALIGN;       # still exists, so it is longer.
		}
		elsif ($next_exon > 0) {
			$hole = TRUE;     # if the next matching exon isn't at the
			last ALIGN;       # front, there's a hole in the alignment
		}

		# Both genes have matching exons at the 5' end. gene1 was
		# hacked at the beginning, gene2 needs hacking.
		shift @g2;

		redo ALIGN;           # keep hacking...
	}

	#----- Process the overall picture
	if    (not $match)          {return 0}
	elsif ($hole)               {return gscOverlap}
	elsif ($long5  and $long3 ) {return gscSuperset}
	elsif ($short5 and $short3) {return gscSubset}
	elsif ($long5  and $short3) {return gscSkew5}
	elsif ($long5  and $short5) {return gscSkew3}
	else                        {return gscIdentical}
}

#----------------------------------------------------------------------
# exonInGene
# Returns the offset of the first "match" to a query exon in sbjct gene.
# Match is defined as a return value of gscIdentical. If the
# gscIncompleteFeatures is used, gscLong5 and gscShort5 are also
# considered a "match".
#----------------------------------------------------------------------
sub _exonInGene {
	my ($mode, $op, $bound, $e1, @gene) = @_;
	my $r; # the result of the search

	SEARCH: for(my $i=0;$i<@gene;$i++) {
		my $e2 = $gene[$i];
		my $r; # the result of the comparison

		#----- Do the comparison using the selected operator
		if ($op == gscCoordinates) {
			$r = $e1->cmpByCoor($e2, $bound);
		}
		elsif ($op == gscGenomic) {
			$r = $e1->cmpByGenomic($e2, $bound);
		}
		elsif ($op == gscScore) {
			$r = $e1->cmpByScore($e2, $bound);
		}
		else {die "ERROR: invalid comparison operator\n"}

		#----- Process the results
		return $i if $r == gscIdentical;
		return $i if $mode == gscIncompleteFeatures and
				($r == gscLong5 or $r == gscShort5);
	}

	# If no matching exons were found, return the undefined value
	return undef;
}


#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

