package Bio::GSC::Metric;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Metric

=head1 DESCRIPTION

Bio::GSC::Metric is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Metric is the base class for sequence metrics. A Metric runs
along the entire length of both strands of a sequence and contains some
kind of scoring at each position (eg. exon probability). Metrics
can be casted to features using the toFeature() method.

=head1 DEPENDENCIES


=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use Bio::GSC::Constants;

#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------
my @Path = split(/:/, $ENV{PERL5LIB});
my %Parser;
foreach my $dir (@Path) {
    my $pdir = "$dir/Bio/GSC/Metric";
	next unless -d $pdir;
	opendir(PD, $pdir) or warn
		"ERROR: can't open modules in $pdir, continuing...";
	PARSER: {
		my $parser = readdir(PD);
		last PARSER unless defined $parser;
		redo PARSER unless $parser =~ /(\w+)\.pm$/;
		redo PARSER if defined $Parser{$1};
		eval "use Bio::GSC::Metric::$1";
		$Parser{$1} = 1;
		redo PARSER;
	}
}

#----------------------------------------------------------------------
# getParserNames
#----------------------------------------------------------------------

=head2 getParserNames()

Description: returns a list of all parsers located in Bio/GSC/Metric.
Looks in the PERL5LIB evironment variable and loads the first parser
of any type found (if there is more than one Bio/GSC/Tool/Xpound in
the PERL5LIB path, only the first is loaded).

 Usage: @names  = Bio::GSC::Metric::getParserNames();

=cut
	  
sub getParserNames {
	return sort keys %Parser;
}

#----------------------------------------------------------------------
# new
#----------------------------------------------------------------------

=head2 new()

Description: instantiates a metric. Note: not usually called on its
own, but by parse(), which creates several metrics from a file.

 Usage: $feature = new Bio::GSC::Feature();

=cut

sub new {
	my ($this, $type, $file) = @_;
	my $m = {};
	bless $m;
	return $m unless defined $type;
	$m->setType($type);
	die "ERROR: file not found [$file]\n" unless -e $file;

	my $parser = "Bio::GSC::Metric::$type" . "::parse";
	$m->$parser($file);
	return $m;
}

#----------------------------------------------------------------------
# Attribute Accessor/Assignment Methods
#----------------------------------------------------------------------

#----------------------------------------------------------------------

=head2 ID

Description: Every metric has a unique identifier.

 Usage: $feature->getID();
        $feature->setID($identifier);

=cut

sub getID {
	 my ($this) = @_;
	 return $this->{'ID'};
}

sub setID {
	 my ($this, $value) = @_;
	 $this->{'ID'} = $value;
}

#----------------------------------------------------------------------

=head2 Type

Description: Most metrics have a Type (eg. Xpound).

 Usage: $feature->getType();
        $feature->setType($name);

=cut

sub getType {
	 my ($this) = @_;
	 return $this->{'Type'};
}

sub setType {
	 my ($this, $value) = @_;
	 $this->{'Type'} = $value;
}

#----------------------------------------------------------------------

=head2 toFeature() needs to be generalized for > < ==

Description:  changes a metric into a list of features. You must
specify the value of the metric for the cutoff.

 Usage: $metric->toFeature(0.75);

=cut

sub toFeature {
	my ($this, $name, $cutoff) = @_;
	my ($p_begin, $p_end, $m_begin, $m_end, $p_score, $m_score);
	for(my $i=0;$i<@{$this->{'plus'}};$i++) {
		my $plus = $this->getValue('plus', $i);
		my $minus = $this->getValue('minus', $i);

		if ($plus > $cutoff) {
			if ($p_score) {$p_end++}
			else {$p_begin = $p_end = $i; $p_score = 1}
		}
		else {
			print "$p_begin  $p_end\n" if $p_score;
			$p_score = 0;
		}

		if ($minus > $cutoff) {
			if ($m_score) {$m_begin++}
			else {$m_end = $m_begin = $i; $m_score = 1}
		}
		else {
			print "$m_begin  $m_end\n" if $m_score;
			$m_score = 0;
		}
	}
}

#----------------------------------------------------------------------
# toTool
#----------------------------------------------------------------------

=head2 toTool()

Description: converts a Metric to a Tool. Works but unfinished.

 Usage: $metric->toTool($how);

=cut

sub toTool {
	my ($metric, $param) = @_;
	my ($operator, $value) = $param =~
			/^value\s+([<>]=*)\s+([\d\.\e]+)/;
	die "ERROR: bad argument\n"
		unless defined $operator and defined $value;
	
	my $tool = new Bio::GSC::Tool;
	$tool->setType($metric->getType());

	my ($pb, $pe, $mb, $me, $ps, $ms);
	for(my $i=0;$i<@{$metric->{'plus'}};$i++) {
		my $plus = $metric->{'plus'}[$i];
		my $minus = $metric->{'minus'}[$i];
		if (eval "$plus $operator $value") {
			if ($ps) {$pe++}
			else {$pb = $pe = $i; $ps = 1}
		}
		else {
			if ($ps) {
				my $feature = new Bio::GSC::Feature;
				$feature->setBegin($pb);
				$feature->setEnd($pe);
				$tool->addFeature($feature);
			}
			$ps = 0;
		}

		if (eval "$minus $operator $value") {
			if ($ms) {$mb++}
			else {$me = $mb = $i; $ms = 1}
		}
		else {
			if ($ms) {
				my $feature = new Bio::GSC::Feature;
				$feature->setBegin($mb);
				$feature->setEnd($me);
				$tool->addFeature($feature);
			}
			$ms = 0;
		}
	}

	return $tool;
}

#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

