package Bio::GSC::NT;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::NT

=head1 SYNOPSIS

<some samples of usage go here>

=head1 DESCRIPTION

Bio::GSC::NT is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA. Bio::GSC::NT has several semi-private methods that
are called from Bio::GSC::Sequence.

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw(
	LoadCodonTable
	complementDNA
	reverseDNA
	antiDNA
	translateDNA
	isORF
	daltonNT);

#----------------------------------------------------------------------
# Class Variables
#----------------------------------------------------------------------
my %Codon = (
	'AAA' => 'K', 'AAC' => 'N', 'AAG' => 'K', 'AAT' => 'N', 
	'ACA' => 'T', 'ACC' => 'T', 'ACG' => 'T', 'ACT' => 'T', 
	'AGA' => 'R', 'AGC' => 'S', 'AGG' => 'R', 'AGT' => 'S', 
	'ATA' => 'I', 'ATC' => 'I', 'ATG' => 'M', 'ATT' => 'I', 
	'CAA' => 'Q', 'CAC' => 'H', 'CAG' => 'Q', 'CAT' => 'H', 
	'CCA' => 'P', 'CCC' => 'P', 'CCG' => 'P', 'CCT' => 'P', 
	'CGA' => 'R', 'CGC' => 'R', 'CGG' => 'R', 'CGT' => 'R', 
	'CTA' => 'L', 'CTC' => 'L', 'CTG' => 'L', 'CTT' => 'L', 
	'GAA' => 'E', 'GAC' => 'D', 'GAG' => 'E', 'GAT' => 'D', 
	'GCA' => 'A', 'GCC' => 'A', 'GCG' => 'A', 'GCT' => 'A', 
	'GGA' => 'G', 'GGC' => 'G', 'GGG' => 'G', 'GGT' => 'G', 
	'GTA' => 'V', 'GTC' => 'V', 'GTG' => 'V', 'GTT' => 'V', 
	'TAA' => '*', 'TAC' => 'Y', 'TAG' => '*', 'TAT' => 'Y', 
	'TCA' => 'S', 'TCC' => 'S', 'TCG' => 'S', 'TCT' => 'S', 
	'TGA' => '*', 'TGC' => 'C', 'TGG' => 'W', 'TGT' => 'C', 
	'TTA' => 'L', 'TTC' => 'F', 'TTG' => 'L', 'TTT' => 'F'
);
my %Dalton = ('A' => 660, 'G' => 660, 'C' => 660, 'T' => 660);

#----------------------------------------------------------------------
# LoadCodonTable
#----------------------------------------------------------------------

=head2 LoadCodonTable()

Description: Loads an alternate codon table. 

 Usage: LoadCodonTable($table_name);

=cut

sub LoadCodonTable {
}

#----------------------------------------------------------------------
# DNA transformations
#----------------------------------------------------------------------

=head2 complementDNA()

Description: Complements a DNA sequence.
 ACGTRYMKWSBDHVN -> TGCAYRKMSWVHDBN

 Usage: $comp = complementDNA($seq);

=cut

sub complementDNA {
	my ($seq) = @_;
	$seq =~ tr[ACGTRYMKWSBDHVN]
			  [TGCAYRKMSWVHDBN];
	return $seq;
}

#----------------------------------------------------------------------
# reverseDNA()
#----------------------------------------------------------------------

=head2 reverseDNA()

Description: Reverses sequence order.
 ACCGGGTTTT -> TTTTGGGCCA

 Usage: $rev = reverseDNA($seq);

=cut

sub reverseDNA {
	my ($seq) = @_;
	$seq = reverse $seq;
	return $seq;
}

#----------------------------------------------------------------------
# antiDNA()
#----------------------------------------------------------------------

=head2 antiDNA()

Description: Gives the antiparallel sequence for a strand of DNA.
 ACCGGGTTTT -> AAAACCCGGT

 Usage: $anti = antiDNA($seq);

=cut

sub antiDNA {
	my ($seq) = @_;
	return reverseDNA(complementDNA($seq));
}

#----------------------------------------------------------------------
# translateDNA()
#----------------------------------------------------------------------

=head2 translateDNA()

Description: Translates a nucleotide sequence to amino acid sequence.
Degenerate/unknown nucleotides are translated to an 'X'. Frame may be
specified as 1, 2, 3, -3, -2, -1 [default is 1]

 Usage: $aa = translateDNA($seq);
        $aa = translateDNA($seq, $frame);

=cut

sub translateDNA {
	my ($dna,$frame) = @_;
	$frame = 1 unless defined $frame;
	if ($frame < 0) {
		$dna = antiDNA($dna);
		$frame = abs($frame);
	}
	my $aa = '';
	for(my $i=$frame-1; $i<length($dna)-(length($dna)%3)-$frame;$i+=3) {
		if (exists $Codon{substr($dna,$i,3)})
			{$aa .= $Codon{substr($dna,$i,3)}}
		else {$aa .= 'X'}
	}
	return $aa;
}

#----------------------------------------------------------------------
# isORF()
#----------------------------------------------------------------------

=head2 isORF()

Description: used to determine if a nucleotide sequence has an open
reading frame. Assumes degenerate nucleotides code for amino acids.

 Usage: $true_falst = isORF($seq);

=cut

sub isORF {
	my ($dna,$frame) = @_;
	my $i;
	my $orf = 1; # starts at true
	if ($frame < 0) {
		$dna = Anti($dna);
		$frame = abs($frame);
	}
	for($i=$frame-1; $i<length($dna)-(length($dna)%3)-$frame;$i+=3) {
		if (    (exists $Codon{substr($dna,$i,3)}) and
				($Codon{substr($dna,$i,3)} eq '*')       ) {
			$orf = 0;
		}
	}
	return $orf;
}

#----------------------------------------------------------------------
# daltonNT
#----------------------------------------------------------------------

=head2 daltonNT()

Description: determines the molecular weight of a nucleotide sequence.
Assumes the sequence is double stranded.

 Usage: $weight = daltonNT($seq);

=cut

sub daltonNT {
}

#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

