package Bio::GSC::ScoringMatrix;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::ScoringMatrix

=head1 SYNOPSIS

<some samples of usage go here>

=head1 DESCRIPTION

Bio::GSC::ScoringMatrix is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::ScoringMatrix offers methods for scoring aligned sequence.
See Bio::GSC::Align for methods of aligning sequences.

=head1 ENVIRONMENT

$MATRIX_HOME is the directory containing matricies.

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();
@EXPORT_OK = qw(
	scorePair
	getAlignment
	LoadMatrix
	AlterMatrix
	ExportMatrix
	PostMatrix);

#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------
my %Matrix;

sub ExportMatrix {return \%Matrix};

#----------------------------------------------------------------------
# PostMatrix
#----------------------------------------------------------------------

=head2 PostMatrix()

Description: Prints the current matrix to STDOUT

 Usage: Bio::GSC::ScoringMatrix::PostMatrix();

=cut

sub PostMatrix {
	my (@char) = @_;
	@char = sort keys %Matrix unless defined @char;
	my @c2 = @char;
	print '    ';
	print (join('   ',@char));
	print "\n";
	foreach my $row (@char) {
		print shift @c2;
		foreach my $col (@char) {
			print (' ' x (4-length($Matrix{$row}{$col}))); 
			print $Matrix{$row}{$col};
		}
		print "\n";
	}
}

#----------------------------------------------------------------------
# LoadMatrix
#----------------------------------------------------------------------

=head2 LoadMatrix()

Description: Loads a new scoring matrix. Matricies are looked for in
the location specified by the MATRIX_HOME environment variable.

 Usage: Bio::GSC::LoadMatrix($name);

=cut

sub LoadMatrix {
	my ($matrix_name) = @_;
	my $matrix_home = $ENV{MATRIX_HOME};

	open(MATRIX,"<$matrix_name")
		or open(MATRIX,"<$matrix_home/$matrix_name")
		or die "can't open $matrix_name";
	while (<MATRIX>) {last if $_ !~ /^#/}
	my @key = split(/\s+/, $_);
	shift @key;
	my @value = <MATRIX>;
	close MATRIX;

	my ($key1, $key2, @mat,$m);
	for(my $i=0;$i<@value;$i++) {
		@mat = split(/\s+/,$value[$i]);
		shift(@mat);
		for(my $k=0;$k<@mat;$k++) {
			$Matrix{$key[$i]}{$key[$k]} = $mat[$k];
		}
	}
	return 1;
}

#----------------------------------------------------------------------
# AlterMatrix
#----------------------------------------------------------------------

=head2 AlterMatrix()

Description: Alters the scores in the current matrix;

 Usage: Bio::GSC::ScoringMatrix::AlterMatrix('F', 'B', 4);

=cut

sub AlterMatrix {
	my ($A, $B, $value) = @_;
	die "any? any?\n" if $A eq 'any' and $B eq 'any';
	my $key;
	if ($A eq 'any')
		{foreach $key (keys %Matrix) {$Matrix{$key}{$B} = $value}}
	elsif ($A eq 'any')
		{foreach $key (keys %Matrix) {$Matrix{$A}{$key} = $value}}
	else {$Matrix{$A}{$B} = $value}
}

#----------------------------------------------------------------------
# scorePair
#----------------------------------------------------------------------

=head2 scorePair()

Description: Scores two aligned sequences (must have same length). The
sequences are passed by reference.

 Usage: $score = Bio::GSC::ScoringMatrix::scorePair($seq1, $seq2);

=cut

sub scorePair {
	my ($seq1, $seq2) = @_;
	my $score = 0;
	for(my $i=0;$i<length($$seq1);$i++) {
		$score += $Matrix{substr($$seq1, $i, 1)}{substr($$seq2, $i, 1)};
	}
	return $score;
}

#----------------------------------------------------------------------
# getAlignment
#----------------------------------------------------------------------

=head2 getAlignment()

Description: takes two aligned sequences and returns an alignment
string. Similar residues (those with positive scores) have a +. The
sequences are passed by reference.

 Usage: $align = Bio::GSC::ScoringMatrix::getAlignment($seq1, $seq2);

=cut

sub getAlignment {
	my ($seq1, $seq2) = @_;
	my $score;
	my $align = '';
	for(my $i=0;$i<length($$seq1);$i++) {
		$score = $Matrix{substr($$seq1, $i, 1)}{substr($$seq2, $i, 1)};
		if (substr($$seq1, $i, 1) eq substr($$seq2, $i, 1))
			{$align .= substr($$seq1, $i, 1)}
		elsif ($score > 0)
			{$align .= '+'}
		else
			{$align .= ' '}
	}
	return ($align);
}

#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTs

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1
