package Bio::GSC::Sequence;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Sequence

=head1 SYNOPSIS

<some samples of usage>

=head1 DESCRIPTION

Bio::GSC::Sequence is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Sequence is the main class used for sequence analysis. Most
users should not have to use the inner calsses (see dependencies
below). A sequence can contain various attributes, including a name,
a unique identifier, a source organism etc. In addition, a sequence may
contain a Metric table and a Tool table. Metrics measure something at
each postion of the sequence. Tools indicate features on the sequence.

=head1 ENVIRONMENT

Optional variables
 MATRIX_HOME
 GETSEQ

=head1 DEPENDENCIES

 Bio::GSC::NT
 Bio::GSC::AA
 Bio::GSC::Utilites
 Bio::GSC::Feature
 Bio::GSC::Tool
 Bio::GSC::Visual

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter Bio::GSC::Visual Bio::GSC::Postscript);
@EXPORT_OK = qw();

use PostData;
use Bio::GSC::Constants;
use Bio::GSC::Tool;
use Bio::GSC::Feature;
use Bio::GSC::Metric;
use Bio::GSC::Visual;
use Bio::GSC::Postscript;
use Bio::GSC::Utilities;
use Bio::GSC::NT;
use Bio::GSC::AA;

#----------------------------------------------------------------------
# Class Variables
#----------------------------------------------------------------------
my @Attribute = qw(
	Name
	Begin
	End
	Length
	Strand
	Score
	P
	Percent
);

my $MaskChar = 'N';
sub setMaskChar {
	($MaskChar) = @_;
}

my $Source = 'file'; # only file supported at this time
sub setSource {
	($Source) = @_;
}

#----------------------------------------------------------------------
# new
#----------------------------------------------------------------------

=head2 new()

Description: Initializes a new sequence object. A sequence object can
be read in from a file, and the type of file may be raw, fasta, genbank,
or embl type (only fasta supported).

 Usage: $seq = new Bio::GSC::Sequence();
        $seq = new Bio::GSC::Sequence(TYPE, $file);

 Types: gscFastaInput
        gscGenbankInput (not supported yet)
        gscEmblInput (not supported yet)
        gscRawInput

=cut

sub new {
	my ($class, $type, $name) = @_;
	my $seq = {};
	bless $seq;
	return $seq unless defined $type;
	if    ($type == gscFastaInput)   {$seq->_newFasta($name)}
	elsif ($type == gscGenbankInput) {$seq->_newGenbank($name)}
	elsif ($type == gscEmblInput)    {$seq->_newEMBL($name)}
	elsif ($type == gscRawInput)     {$seq->_newRaw($name)}
	else  {die "ERROR unknown type: $type\n"}
	return $seq;
}

#----------------------------------------------------------------------
# _newRaw
#----------------------------------------------------------------------
sub _newRaw {
	my ($this, $name) = @_;
	my $seq = "";
	open(IN, "<$name") or die "ERROR: file not found [$name]\n";
	while(<IN>) {
		$_ =~ s/\s+//g;
		$seq .= $_;
	}
	close IN;
	$this->setSequence($seq);
}

#----------------------------------------------------------------------
# _newFasta
#----------------------------------------------------------------------
sub _newFasta {
	my ($this, $name) = @_;
	my ($def, $seq);
	my $data;
	if ($Source eq 'file')
		{($def, $seq) = Bio::GSC::Utilities::getFasta($name)}
	elsif ($Source eq 'external')
		{($def, $seq) = Bio::GSC::Utilities::fetchFasta($name)}
	else {die "unrecognized source: $Source (file or external)\n"}
	($name) = $def =~ /^>(\w+)/;
	$this->setDefinition($def);
	$this->setName($name);
	$this->setSequence($seq);
	return $data;
}

#----------------------------------------------------------------------
# _newGenbank
#----------------------------------------------------------------------
sub _newGenbank {
	my ($this, $name) = @_;
	my $data;
	if ($Source eq 'file') {}          # need to parse report
	elsif ($Source eq 'external') {}   # need to parse report
	else {die}
	return $data;
}

#----------------------------------------------------------------------
# _newEMBL
#----------------------------------------------------------------------
sub _newEMBL {
	my ($this, $name) = @_;
	my $data;
	if ($Source eq 'file') {}          # need to parse report
	elsif ($Source eq 'external') {}   # need to parse report
	else {die}
	return $data;
}

#----------------------------------------------------------------------
# getDefinition
#----------------------------------------------------------------------

=head2 getDefinition()

Description: Returns a fasta definition line for a sequence object.
If a fasta file was not used to initialize the object, a definition
line is created using other attributes.

 Usage: $defline = $seq->getDefinition();

=cut

sub getDefinition {
	my ($this) = @_;
	if (exists $this->{'Definition'}) {return $this->{'Definition'}}
	else {return _createDef($this)}
}

#----------------------------------------------------------------------
# setDefinition
#----------------------------------------------------------------------

=head2 setDefinition()

Description: Assigns a new fasta-style definition line to a sequence
object.

 Usage: $seq->setDefinition($fasta_def_line);

=cut

sub setDefinition {
	my ($this, $def) = @_;
	$this->{'Definition'} = $def;
	return;
}

#----------------------------------------------------------------------
# _createDef
#----------------------------------------------------------------------

=head2 _createDef()

Description: Creates a fasta-style definition line from various
object attributes. Useful for Genbank -> Fasta. Called by getDef().

 Usage: my $def = _createDef()

=cut

sub _createDef {
	my ($this) = @_;
	my $def = getID() . ' '; 
	$def .= getName() . ' ';
	$def .= '[' . getOrganism() . '] ';
	$def .= getDescription();
	return $def;
}

#----------------------------------------------------------------------
# getID
#----------------------------------------------------------------------

=head2 getID()

Description: Accessor method for a unique identifier. This should be
the gi number but may be any identifier.

 Usage: $seq->getID();

=cut

sub getID {
	my ($this) = @_;
	return $this->{'ID'}
}

#----------------------------------------------------------------------
# setID
#----------------------------------------------------------------------

=head2 setID()

Description: Assigns an identifier string to a sequence object.

 Usage: $seq->setID($string);

=cut

sub setID {
	my ($this, $id) = @_;
	$this->{'ID'} = $id;
	return;
}

#----------------------------------------------------------------------
# getName
#----------------------------------------------------------------------

=head2 getName()

Description: Accessor method to get the name of a sequence.

 Usage: $seq->getID();

=cut

sub getName {
	my ($this) = @_;
	return $this->{'Name'};
}

#----------------------------------------------------------------------
# setName
#----------------------------------------------------------------------

=head2 setName()

Description: Assigns a name to a sequence object.

 Usage: $seq->setID($string);

=cut

sub setName {
	my ($this, $name) = @_;
	$this->{'Name'} = $name;
	return;
}

#----------------------------------------------------------------------
# getOrganism
#----------------------------------------------------------------------

=head2 getOrganism()

Description: Accessor method to get the organism name.

 Usage: $seq->getOrganism();

=cut

sub getOrganism {
	my ($this) = @_;
	return $this->{'Organism'};
}

#----------------------------------------------------------------------
# setOrganism
#----------------------------------------------------------------------

=head2 setOrganism()

Description: Assigns an organism name to a sequence object.

 Usage: $seq->setOrganism($organism);

=cut

sub setOrganism {
	my ($this, $org) = @_;
	$this->{'Organism'} = $org;
	return;
}

#----------------------------------------------------------------------
# getSequence
#----------------------------------------------------------------------

=head2 getSequence()

Description: Returns the sequence or a slice of the sequence. No
arguments are necessary. If given from and to coordinates, a slice of
the sequence is returned. If from is greater than to, the reverse
complement is returned. You must always specify from and to as a pair.
You may also specify the strand as 'reverse' 'complement' or
'antiparallel' (actually 'r' 'c' or 'a' is enough).

 Usage: $seq->getSequence();
        $seq->getSequence($from, $to);
        $seq->getSequence($from, $to, $strand);

=cut

sub getSequence {
	my ($this, @arg) = @_;
	my ($from, $to, $strand, $seq);
	if    (@arg == 1) {($strand) = @arg}
	elsif (@arg == 2) {($from, $to) = @arg}
	elsif (@arg == 3) {($from, $to, $strand) = @arg}

	#----- get the sequence
	if (defined $from) {
		my ($offset, $length);
		$offset = $from < $to ? $from : $to;
		$length = abs($from - $to);
		$seq = substr($this->{'Sequence'}, $offset, $length);
		$seq = $from > $to ?
			Bio::GSC::NT::antiDNA($seq) : $seq;
	}
	else {$seq = $this->{'Sequence'}}

	#----- transform sequence depending upon strand argument
	if (not defined $strand) {} # that's okay
	elsif ($strand =~ /^c/)
		{$seq = Bio::GSC::NT::complementDNA($seq)}
	elsif ($strand =~ /^r/)
		{$seq = Bio::GSC::NT::reverseDNA($seq)}
	elsif ($strand =~ /^a/)
		{$seq = Bio::GSC::NT::antiDNA($seq)}
	return $seq;
}

#----------------------------------------------------------------------
# setSequence
#----------------------------------------------------------------------

=head2 setSequence()

Description: Assigns a sequence to a sequence object.

 Usage: $seq->setSequence($sequence);

=cut

sub setSequence {
	my ($this, $seq) = @_;
	$this->{'Sequence'} = $seq;
	return;
}

#----------------------------------------------------------------------
# getLength
#----------------------------------------------------------------------

=head2 getLength()

Description: Retrieves the length of the sequence object.

 Usage: $seq->getLength();

=cut

sub getLength {
	my ($this) = @_;
	return length($this->{'Sequence'});
}

#----------------------------------------------------------------------
# addMetric
#----------------------------------------------------------------------

=head2 addMetric()

Description: Adds a metric markup to a sequence object. A metric runs
along both strands of a sequence and measures some value at each
position. A metric might be exon probability or hydropathy (if this was
an amino acid sequence, which it isn't). See Bio::GSC::Metric for more
information. addMetric requires three arguments, a unique identfier
used as a key to the Metric table, a type of metric (eg. Xnu), and a
file name.

 Usage: $seq->addMetric($name, $type, $file);

=cut

sub addMetric {
	my ($this, $id, $type, $file) = @_;
	my $tool = new Bio::GSC::Metric($type, $file);
	$this->{'Metric'}{$id} = $tool;
}

#----------------------------------------------------------------------
# addTool
#----------------------------------------------------------------------

=head2 addTool()

Description: Adds a Tool to a sequence object. Tools mark up a
sequence using discrete coordinates called Features. A Feature always
contains Begin and End attributes but may contain others such Score or
P. See Bio::GSC::Tool and Bio::GSC::Feature for more information. Most
sequence markups, like database similarities and gene predictions are
markup Tools. Tools are usually loaded from a file created by the
output of a program. addTool() requires three arguments and takes
additional optional filtering arguments. The first argument is a unique
name used for the key in the Tool table, the second argument is the
type of program (really the type of parser - all WUBLAST programs are
the same type), and the third argument is the file name. Filtering
parameters may also be given. Since Features may contain SubFeatures
(eg. genes are features that contain exon subfeatures), multiple
filtering parameters may be given, and in this case the first filter
applies to the feature, and the second to the subfeatures (and so on).

 Usage: $seq->addTool($name, $type, $file);
        $seq->addTool($name, $type, $file, 'Score > 50');
        $seq->addTool($name, $type, $file, 'Score > 50', 'Score > 9');

=cut

sub addTool {
	my ($this, $id, $type, $file, @filter) = @_;
	my $tool = new Bio::GSC::Tool($type, $file, @filter);
	$this->{'Tool'}{$id} = $tool;
}

#----------------------------------------------------------------------
# copyTool
#----------------------------------------------------------------------

=head2 copyTool()

Description: Makes a copy of one of the Tool in the Tool table. You
must give the name of the Tool to copy, and a new unique name.

 Usage: $seq->duplicateTool($name, $new_name);

=cut

sub copyTool {
	my ($this, $name1, $name2) = @_;
	my $template = $this->getTool($name1);
	my $copy = $template->clone();
	$this->{'Tool'}{$name2} = $copy;
}

#----------------------------------------------------------------------
# removeTool
#----------------------------------------------------------------------

=head2 removeTool()

Description: Removes a Tool from a sequence object.

 Usage: $seq->removeTool($name);

=cut

sub removeTool {
	my ($this, $id) = @_;
	undef($this->{'Tool'}{$id});
}

#----------------------------------------------------------------------
# filterTool
#----------------------------------------------------------------------

=head2 filterTool()

Description : Allows filtering of a Tool's features based on any
feature attribute(s). See Bio::GSC::Feature::queryFeature for more
information.

 Usage: $seq->filterTool($tool_name, $filtering_string);

=cut

sub filterTool {
	my ($this, $id, @filter) = @_;
	my $tool = $this->getTool($id);
	$tool->filterFeatures(@filter);
	$this->{'Tool'}{$id} = $tool;
}

#----------------------------------------------------------------------
# getTool
#----------------------------------------------------------------------

=head2 getTool()

Description: Gets a Tool from a sequence object. Used internally.

 Usage: $seq->getTool($name);

=cut

sub getTool {
	my ($this, $id) = @_;
	return $this->{'Tool'}{$id};
}

#----------------------------------------------------------------------
# getMetric
#----------------------------------------------------------------------

=head2 getMetric()

Description : Gets a metric from a sequence object. Used internally.

 Usage: $seq->getMetric($name);

=cut

sub getMetric {
	my ($this, $id) = @_;
	return $this->{'Metric'}{$id};
}

#----------------------------------------------------------------------
# joinTool
#----------------------------------------------------------------------

=head2 joinTool()

Description: Creates a new Tool by joining two others.
See Bio::GSC::Tool::Join() for more information.

 Usage: $seq->joinTool($newtool, $tool1, $tool2);

=cut

sub joinTool {
	my ($this, $id, $tool1, $tool2) = @_;
	my $tool = Bio::GSC::Join($id, $tool1, $tool2);
	$this->{'Tool'}{$id} = $tool;
}

#----------------------------------------------------------------------
# intersectTool
#----------------------------------------------------------------------

=head2 intersectTool()

Description: Creates a new Tool using the intersection of two others.
See Bio::GSC::Tool::Intersect() for more information.

 Usage: $seq->intersectTool($name, $tool1, $tool2, MODE,
                              OPERATOR, $boundary);

=cut

sub intersectTool {
	my ($this, $id, $tool1, $tool2, $mode, $operator, $bound) = @_;
	my $tool = Bio::GSC::Tool::Intersect(
		$this->getTool($tool1), $this->getTool($tool2),
		$mode, $operator, $bound);
	$this->{'Tool'}{$id} = $tool;
}

#----------------------------------------------------------------------
# excludeTool
#----------------------------------------------------------------------

=head2 excludeTool()

Description: Creates a new Tool excluding the intersection of the two
compared. See Bio::GSC::Tool::Exclude() for more information.

 Usage: $seq->excludeTool($name, $tool1, $tool2, MODE,
                            OPERATOR, $boundary);

=cut

sub excludeTool {
	my ($this, $id, $tool1, $tool2, $mode, $operator, $bound) = @_;
	my $tool = Bio::GSC::Tool::Exclude(
		$this->getTool($tool1), $this->getTool($tool2),
		$mode, $operator, $bound);
	$this->{'Tool'}{$id} = $tool;
}

#----------------------------------------------------------------------
# clusterTool
#----------------------------------------------------------------------

=head2 clusterTool()

Description: Trims the feature table of a Tool by removing similar
entries. See Bio::GSC::Tool::clusterFeatures() for more information.

 Usage: $seq->clusterTool($name, MODE, OPERATOR, $boundary);

=cut

sub clusterTool {
	my ($this, $id, $mode, $op, $bound) = @_;
	my $m = $this->getTool($id);
	$m->clusterFeatures($mode, $op, $bound);
}

#----------------------------------------------------------------------
# maskSequence
#----------------------------------------------------------------------

=head2 maskSequence()

Description: uses Tools to mask regions of the sequence.

 Usage: my $mask = $seq->maskSequence($edge, @tool);

=cut

sub maskSequence {
	my ($this, $edge, @tool) = @_;
	my $seq = $this->getSequence;
	my $len = $this->getLength;
	$seq = uc($seq);
	foreach my $id (@tool) {
		my $tool = $this->getTool($id);
		my $feature = $tool->getFeature;
		foreach my $f (@$feature) {
			_maskloop($f, \$seq, $edge, $len);
		}
	}
	$seq =~ s/[a-z]/$MaskChar/g;
	$seq = uc($seq);
	return \$seq;
}

#----------------------------------------------------------------------
# fenestrateSequence
#----------------------------------------------------------------------

=head2 fenestrateSequence()

Description: Uses Tools to reverse-mask regions of the sequence.

 Usage: my $window = $seq->fenestrateSequence($edge, @tool);

=cut

sub fenestrateSequence {
	my ($this, $edge, @tool) = @_;
	my $seq = $this->getSequence;
	my $len = $this->getLength;
	$seq = uc($seq);
	foreach my $id (@tool) {
		my $tool = $this->getTool($id);
		my $feature = $tool->getFeature;
		foreach my $f (@$feature) {
			_maskloop($f, \$seq, $edge, $len);
		}
	}
	$seq =~ s/[A-Z]/$MaskChar/g;
	$seq = uc($seq);
	return \$seq;
}

#----------------------------------------------------------------------
# _maskloop
#---------------------------------------------------------------------
sub _maskloop {
	my ($f, $seq_ref, $edge, $len) = @_;
	if ($f->isComplex) {
		my $sub = $f->getSubFeature;
		foreach my $s (@$sub) {_maskloop($s, $seq_ref, $edge, $len)}
	}
	else {
		my ($start, $length);
		if ($f->getStrand > 0)
			{($start, $length) = ($f->getBegin, $f->getSpan)}
		else
			{($start, $length) = ($f->getEnd, $f->getSpan)}

		my $slice;
		if ($edge) {
			$start = $start - $edge;
			$start = 0 if $start < 1;

			if ($start + $length + ($edge * 2) > $len)
				{$length = $len - $start}
			else
				{$length += ($edge * 2)}
		}
		$slice = substr($$seq_ref, $start, $length);
		$slice = lc($slice);
		substr($$seq_ref, $start, $length) = $slice;
	}
}

#----------------------------------------------------------------------
# dumpTool
#----------------------------------------------------------------------

=head2 dumpTool()

Description: Prints out Tool as a table (currently just simple).

 Usage: $seq->dumpTool($tool_name, @attributes);

=cut

sub dumpTool {
	my ($this, $id, @field) = @_;
	my $fl = 10; # field length
	print "$id\n";
	@field = @Attribute unless defined @field;
	foreach my $field (@field)
		{print " " x ($fl - length($field)), $field}
	print "\n";
	foreach my $feature (@{$this->{'Tool'}{$id}{'FeatureTable'}}) {
		foreach my $att (@field) {
			my $string = eval "\$feature->get$att()";
			$string = "" unless defined $string;
			print " " x ($fl - length($string)), $string;
		}
	print "\n";
	}
}

#----------------------------------------------------------------------
# metricToTool
#----------------------------------------------------------------------

=head2 metricToTool()

Description: converts a Metric to a Tool. Partially finished.

 Usage: $seq->metricToTool($old_name, $new_name, $how);

=cut

sub metricToTool {
	my ($this, $name, $new, $param) = @_;
	my $metric = $this->getMetric($name);
	my $tool = $metric->toTool($param);
	$this->{'Tool'}{$new} = $tool;
}

#----------------------------------------------------------------------
# serialize
#----------------------------------------------------------------------

=head2 serialize()

Description: serializes the sequence object.

 Usage: $seq->serialize();
		$seq->serialize($filename);

=cut

sub serialize {
	my ($this, $file) = @_;
	Bio::GSC::Utilities::serialize($this);
}

#----------------------------------------------------------------------
# acedbDump
#----------------------------------------------------------------------

=head2 acedbDump()

Description: Dumps the entire sequence object into an acedb flatfile.
You must give a PARAMETER specifying how the dumping will occur for each
Tool (metrics are not supported). You must also supply an ACEDB display
method name. There is also an optional prefix for efetching. Note that
Subsequence dumps only work with complex features (any simple features
will simply be skipped). Also, any complex features are automatically
expanded.


 Usage: $seq->acedbDump($name, PARAMETER, $display, $prefix);

   Parameter                   Behavior of Features
   ----------------------      ----------------------------------------
   gscACEdumpFeature           become ACEDB Features
   gscACEdumpDNA_Homol         become DNA_Homols
   gscACEdumpPEP_Homol         become PEP_Homols
   gscACEdumpSubsequence       become Subsequences

=cut

sub acedbDump {
	my ($this, $id, $param, $display, $prefix) = @_;
	my $tool = $this->getTool($id);
	my $name = $this->getName($id);
	if    ($param == gscACEdumpFeature) {
		_toFeature($tool, $name, $display)}
	elsif ($param == gscACEdumpDNA_Homol) {
		_toHomol($tool, $param, $name, $display, $prefix)}
	elsif ($param == gscACEdumpPEP_Homol) {
		_toHomol($tool, $param, $name, $display, $prefix)}
	elsif ($param == gscACEdumpSubsequence) {
		_toSubsequence($tool, $name)}
	elsif ($param == gscACEdumpMatchTable) {
		_toMatchTable($tool, $name, $display, $prefix)}
	else  {die "ERROR: unsupported parameter\n"}
}

###############################################
# some private functions for dumping to acedb #
###############################################
sub _toMatchTable {
	my ($tool, $name, $display, $prefix) = @_;
	print "\nMatchTable Sequence-$name\n";
	my $feature = $tool->getFeature();
	foreach my $f (@$feature) {_printMatchTable($f, $display, $prefix)}
}
sub _printMatchTable {
	my ($f, $display, $prefix, $id) = @_;
	unless (defined $id) {
		$id = $f->getSbjctDef;
		$id = substr($id, 1, index($id, " "));
	}

	if ($f->isComplex) {
		my $subfeature = $f->getSubFeature;
		foreach my $sf (@$subfeature)
			{_printMatchTable($sf, $display, $prefix, $id)}
	}
	else {
		my $sc = $f->getScore;
		my $qb = $f->getBegin;
		my $ql = $f->getLength;
		my $qe = $f->getEnd;
		my $qs = $f->getStrand;

		my $sb = $f->getSbjctBegin;
		my $sl = $f->getSbjctLength;
		my $se = $f->getSbjctEnd;
		my $ss = $f->getSbjctStrand;

		my $qgap = $f->getQueryGap;
		my $sgap = $f->getSbjctGap;

		my $qseq = $f->getQuerySeq;
		my $sseq = $f->getSbjctSeq;

		#----- must check if the alignment is to protein -----#
		my $protein = FALSE;
		if ($ql > $sl * 2) {
			$protein = TRUE;
			$ql = ($ql) / 3;
		}

		if (defined %$qgap or defined %$sgap) {
			my $shatter = _shatter($qgap, $sgap, $ql, $sl, $qs, $ss);
			print "\n__________________SHATTER START________________\n";
			print "$qseq $ql\n";
			print "$sseq $sl\n";
			foreach my $frag (@$shatter) {
				my ($_qb, $_qe, $_se, $_sb);

=foo

				if ($f->getStrand > 0) {
					$_qb = $qb + $frag->{'qbegin'} - 1;
					$_qe = $qb + $frag->{'qend'} - 1;
				}
				else {
					$_qb = $qe - $frag->{'qbegin'} + 1;
					$_qe = $qe - $frag->{'qend'} + 1;
				}
				if ($f->getSbjctStrand > 0) {
					$_sb = $sb + $frag->{'sbegin'} - 1;
					$_se = $sb + $frag->{'send'} - 1;
				}
				else {
					$_sb = $se - $frag->{'sbegin'} + 1;
					$_se = $se - $frag->{'send'} + 1;
				}

=cut

				$_qb = $frag->{'qbegin'};
				$_qe = $frag->{'qend'};
				$_sb = $frag->{'sbegin'};
				$_se = $frag->{'send'};

				print ">Sequence $prefix:$id $display $sc ";
				print "$_qb $_qe $_sb $_se\n";
			}
			print "__________________SHATTER END__________________\n\n";
		}
		else {
			print "Sequence $prefix:$id $display $sc $qb $qe $sb $se\n";
		}
	}
}
###############################################
# _shatter                                    #
#                                             #
# takes gap hashes and makes a list of coords #
#   @list                                     #
#      {qbegin}                               #
#      {qend}                                 #
#      {sbegin}                               #
#      {send}                                 #
###############################################
sub _shatter {
	my ($qgap, $sgap, $q_last, $s_last, $q_strand, $s_strand) = @_;
	my @coor;
	my @q_start = sort {$a<=>$b} keys %$qgap;
	my @s_start = sort {$a<=>$b} keys %$sgap; 

	my ($qb, $qe, $sb, $se) = (1, 1, 1, 1);
	my ($qb_next, $sb_next);
	my ($q_seen, $s_seen) = (0, 0); # number of gaps seen so far
	my $redo = TRUE;
	FIND_GAP: {
		if (not defined $q_start[0] and not defined $s_start[0]) {
			$redo = FALSE;
print "l>";
		}
		elsif (    (defined $q_start[0] and not defined $s_start[0])
				or (        $q_start[0]  <  $s_start[0])           ) {
print "q>";

			$qe = shift @q_start;
			$se = $qe;
			$qb_next = $qe + 1;
			$sb_next = $se + $qgap->{$qe} + $q_seen + 1;
			$q_seen += $qgap->{$qe};
		}
		elsif (    (defined $s_start[0] and not defined $q_start[0])
				or (        $s_start[0]  <  $q_start[0])           ) {
print "s>";

			$se = shift @s_start;
			$qe = $se;
			$sb_next = $se + 1;
			$qb_next = $qe + $sgap->{$se} + $s_seen + 1;
			$s_seen += $sgap->{$se};
		}
		else {warn "Double gaps!? Garbage in garbage out...\n"}

		my %fragment = (
			'qbegin' => $qb,
			'qend'   => $qe,
			'sbegin' => $sb,
			'send'   => $se);
		push @coor, \%fragment unless $redo == FALSE;

print "QB$qb QE$qe SB$sb SE$se QS$q_seen SS$s_seen\n";
		$qb = $qb_next;
		$sb = $sb_next;
		redo FIND_GAP if $redo == TRUE;
	}
	$qe = $q_last;
	$se = $s_last;
	my %fragment = (
		'qbegin' => $qb,
		'qend'   => $qe,
		'sbegin' => $sb,
		'send'   => $se);
	push @coor, \%fragment;
	return \@coor;
}
sub _toFeature {
	my ($tool, $name, $display) = @_;
	print "\nSequence $name\n";
	my $feature = $tool->getFeature();
	foreach my $f (@$feature) {_printFeature($f, $display)}
}
sub _printFeature {
	my ($f, $display) = @_;
	if ($f->isComplex) {
		my $subfeature = $f->getSubFeature;
		foreach my $sf (@$subfeature) {_printFeature($sf, $display)}
	}
	else {
		my $id = $f->getName();
		my $sc = $f->getScore();
		my $qb = $f->getBegin();
		my $qe = $f->getEnd();
		print "Feature $display $qb $qe $sc \"$id\"\n";
	}
}
sub _toHomol {
	my ($tool, $param, $name, $display, $prefix) = @_;
	my $homol = $param == gscACEdumpDNA_Homol
		? 'DNA_homol'
		: 'Pep_homol';
	print "\nSequence $name\n";
	my $feature = $tool->getFeature();
}
sub _toSubsequence {
	my ($tool, $name) = @_;
	my $feature = $tool->getFeature();
	foreach my $f (@$feature) {
		next unless $f->isComplex();
		my $id  = $f->getSbjctID;
		my $begin = $f->getBegin;
		my $end   = $f->getEnd;
		print "\nSequence $name\n";
		print "Subsequence $name.$id $begin $end\n";
		print "\nSequence $name.$id\n";
		my $sf = $f->getSubFeature();
		foreach my $s (@$sf) {
			my $b = $s->getBegin;
			my $e = $s->getEnd;
			if ($s->getStrand > 0) {
				$b = $b - $begin + 1;
				$e = $e - $begin + 1;
			}
			else {
				$b = $begin - $b + 1;
				$e = $begin - $e + 1;
			}
			print "Source_Exons $b $e\n";
		}
	}
}

#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

