package Bio::GSC::Tool;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Tool

=head1 DESCRIPTION

Bio::GSC::Tool is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Tool is the base class for a sequence markup Tool. A
Tool is defined as an algorithm that creates sequence features The
following programs are supported:
 Blast        - Local alignment search tool (Gish)
 Dust         - Complexity filter (Tatusov & Lipman)
 Genscan      - Gene prediction (Burge)
 Inverted     - Inverted repeat finder (Durbin & Meig)
 Quicktandem  - Tandem repeat finder (Durbin & Meig)
 RepeatMasker - Finds known repeat families (Smit)
 Seg          - Complexity filter (Wooton & Federhen)
 Xnu          - Repeat filter (Claverie & States)
 Xpound_IK    - Exon metric (Myriad Genetics)

=head1 DEPENDENCIES

 Bio::GSC::Feature
 Bio::GSC::Metric
 Bio::GSC::Utilities

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use PostData;
use Bio::GSC::Constants;
use Bio::GSC::ScoringMatrix;
use Bio::GSC::Feature;
use Bio::GSC::Utilities qw(getFasta);


#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------

#--------------- Autoload all Parsers in Bio/GSC/Tool -----------------
my @Path = split(/:/, $ENV{PERL5LIB});
my %Parser;
foreach my $dir (@Path) {
	my $pdir = "$dir/Bio/GSC/Tool";
	next unless -d $pdir;
	opendir(PD, $pdir) or warn
		"ERROR: can't open modules in $pdir, continuing...";
	PARSER: {
		my $parser = readdir(PD);
		last PARSER unless defined $parser;
		redo PARSER unless $parser =~ /(\w+)\.pm$/;
		redo PARSER if defined $Parser{$1};
		eval "use Bio::GSC::Tool::$1";
		$Parser{$1} = 1;
		redo PARSER;
	}
}

my $Trash; # holds a ref to the last list of Intersect or Exclude

#----------------------------------------------------------------------
# new
#----------------------------------------------------------------------

=head2 new()

Description: instantiates a Tool. A Tool creates adds an entry to
the Tool table using the output of a program (eg. a gene finder). You
must specify the an identfier, a type, a file location, and an optional
filtering statement. The identifier must be a unique name in the Tool
table. The type must be one of the available parsers. To get the names
of all the parsers use Bio::GSC::Tool::getParserNames (see below). The
filtering statement uses Bio::GSC::Feature::queryFeature. See the
corresponding module for more information.

 Usage: $tool = new Bio::GSC::Tool($id, $type, $file, $filter);

=cut

sub new {
	my ($m, $type, $file, @filter) = @_;
	$m = {};
	bless $m;
	return $m unless defined $type;

	$m->setType($type);
	die "ERROR: file not found: $file\n" unless -e $file;

	my $parser = "Bio::GSC::Tool::$type" . "::parse";
	$m->$parser($file, @filter);
	return $m;
}

#----------------------------------------------------------------------
# clone
#----------------------------------------------------------------------

=head2 clone()

Description: Dupliates a Tool.

 Usage: $duplicate = $tool->clone();

=cut

sub clone {
	my ($this) = @_;
	my $m = new;
	foreach my $field (keys %$this) {
		$m->{$field} = $this->{$field} unless ref($this->{$field});
	}
	my $features = $this->getFeature();
	foreach my $f (@$features) {
		my $c = $f->clone();
		$m->addFeature($c);
	}
	return $m;
}

#----------------------------------------------------------------------
# getParserNames
#----------------------------------------------------------------------

=head2 getParserNames()

Description: returns a list of all parsers located in Bio/GSC/Tool. 
Looks in the PERL5LIB evironment variable and loads the first parser
of any type found (if there is more than one Bio/GSC/Tool/Blast in
the PERL5LIB path, only the first is loaded).

 Usage: @names  = Bio::GSC::Tool::getParserNames();

=cut

sub getParserNames {
	return sort keys %Parser;
}

#----------------------------------------------------------------------
# Attribute Accessor/Assignment Methods
#----------------------------------------------------------------------

#----------------------------------------------------------------------
# Type
#----------------------------------------------------------------------

=head2 Type Attribute

Description: most Tools have a Type. A Type is determined by the kind
of output file parsed. For example, all Blast programs are of the
Blast type. Tools that don't have a type would be "hand-made" Tools
that don't come from the output of a program. Type is usually assigned
when adding tools.

 Usage: $tool->getType();
        $tool->setType($type);

=cut

sub getType {
	my ($this) = @_;
	return $this->{'type'};
}
sub setType {
	my ($this, $value) = @_;
	$this->{'type'} = $value;
}

#----------------------------------------------------------------------
# Feature
#----------------------------------------------------------------------

=head2 Feature Attribute

Description : a Feature is a markup of the sequence. The Feature
attribute is really a reference to a list (ARRAY) of features. 
getFeature returns a reference to the list, while setFeature assigns
a list reference. addFeature pushes features onto the list.
getFlatFeature returns a list reference, but expands all complex
features.

 Usage: $feature->getFeature();
        $feature->setFeature(\@list_of_features);
        $feature->addFeature($feature);
		$feature->getFlatFeature();

=cut

sub getFeature {
	my ($this) = @_;
	return $this->{'FeatureTable'};
	# should return by sorting?
}

sub setFeature {
	my ($this, $value) = @_;
	die "not an ARRAY reference\n" unless ref $value eq 'ARRAY';
	$this->{'FeatureTable'} = $value;
}

sub addFeature {
	my ($this, $value) = @_;
	push(@{$this->{'FeatureTable'}}, $value);
}

sub getFlatFeature {
	my ($this) = @_;
	my @flatlist;
	foreach my $f (@{$this->{'FeatureTable'}}) {
		if ($f->isComplex()) {
			#----- only going one level deep right now
			my $sublist = $f->getSubFeature();
			push(@flatlist, @$sublist);
		}
		else {push(@flatlist, $f)}
	}
	return \@flatlist;
}


#----------------------------------------------------------------------
# sortFeatures
#----------------------------------------------------------------------

=head2 sortFeatures()

Description: Sorts a list of features. You must specify a sorting
operator.

 Operator          Comments
 --------          --------
 gscSortGenomic    sorts by strand (plus first), then by Begin
 gscSortScore      sorts numbers in descending order
 gscSortP          sorts numbers is ascending ourder

 Usage: $feature->sortFeatures(OPERATOR);

=cut

sub sortFeatures {
	my ($f, $op) = @_;
	if    ($op == gscSortGenomic) {@$f = sort _genomicSort(@$f)}
	elsif ($op == gscSortScore)   {@$f = sort _scoreSort(@$f)}
	elsif ($op == gscSortP)       {@$f = sort _pSort(@$f)}
	return $f;
}

# Some private sorting functions #
sub _genomicSort {
	return
		$b->getStrand() <=> $a->getStrand or
		$a->getBegin()  <=> $b->getBegin  or
		$a->getBegin()  <=> $b->getEnd;
}
sub _scoreSort {
	return $b->getScore() <=> $a->getScore();
}
sub _pSort {
	return
		$a->getP()       <=> $b->getP() or
		$a->getPercent() <=> $b->getPercent();
}

#----------------------------------------------------------------------
# filterFeatures
#----------------------------------------------------------------------

=head2 filterFeatures()

Description: allows you to filter (remove) sequence features based on
any of their values (or combination of values). This can be useful if
a Tool contains many features of little worth. Filtering can also
take place during parsing. See Bio::GSC::Feature::queryFeature for query
syntax. You may wish to make a duplicate of the Features beforehand.

 Usage: $feature->filterFeatures($query);

=cut

sub filterFeatures {
	my ($this, @filter) = @_;
	my @keep;
	my $f = $this->getFeature();
	return unless defined @$f and @$f;
	for(my $i=0;$i<@$f;$i++) {
		push(@keep, $f->[$i]) if $f->[$i]->queryFeature(@filter);
	}
	$this->setFeature(\@keep);
}


#----------------------------------------------------------------------
# clusterFeatures
#----------------------------------------------------------------------

=head2 clusterFeatures()

Description: Trims the features of a Tool by removing similar
entries. You must choose a mode, operator, and boundary for the
comparisons. There are five modes and three operators (see below) You
must also specify a  boundary for the comparison. This interger
value determines how close the features must be to each other. Using
the coorOperator, it is the difference between Begin and End attributes.
Using the seqOperator, it is the amount of extra sequence grabbed from
the genomic. Using the scoreOperator, it is the cutoff score for the 
comparison to be "true". 
    
 Mode                    Behavior
 ---------------------   -----------------------------------------------
 gscSimpleFeatures        complex features are treated as simple
 gscFlattenedFeatures     complex features are converted to simple
 gscIdenticalFeatures     all subfeatures must be similar 
 gscIncompleteFeatures    contiguous subfeatues must be similar 
 gscOverlappingFeatures   some subfeatures must be similar 
    
 Operator        Behavior
 --------------  -------------------------------------------------
 gscCoordinates  the Begin and End of each feature is compared
 gscGenomic      the genomic sequence is used for the comparison
 gscScore        the sbjct (or genomic) is used for the comparison


 Usage: $tool->clusterFeatures(MODE, OPERATOR, $boundary);

=cut

sub clusterFeatures {
	my ($m, $mode, $op, $bound) = @_;
	my $f;
	if ($mode == gscFlattenedFeatures) {$f = $m->getFlatFeature}
	else {$f = $m->getFeature()}

	return unless defined @$f; # no sense clustering nothing
	return unless @$f > 1;     # no sense clustering just one
	my @pool = @$f;            # pool of individual features
	my @redundant;
	my @unique;
	POOL: while(@pool) {
		my $query = shift @pool;

		SEARCH: for(my $i=0;$i<@pool;$i++) {
			my $cmp = $query->cmpFeature($pool[$i], $mode, $op, $bound);
			next SEARCH if $cmp == FALSE;

			# Jump out immediately (greedy) if you make a match
			if ($cmp == gscIdentical) {
				push @redundant, splice(@pool, $i, 1);
				unshift @pool, $query;
				redo POOL;
			}
			# Jump out on any match if overlapping features are okay
			if ($mode == gscOverlappingFeatures) {
				push @redundant, splice(@pool, $i, 1);
				unshift @pool, $query;
				redo POOL;
			}
			# Take the longest of the incomplete features
			if ($mode == gscIncompleteFeatures) {
				if ($cmp == gscSuperset or $cmp == gscLong5 or
						$cmp == gscLong3 or $cmp == gscSkew5) {
					push @redundant, splice(@pool, $i, 1);
					unshift @pool, $query;
					redo POOL;
				}
				elsif ($cmp == gscSubset or $cmp == gscShort5 or
						$cmp == gscShort3 or $cmp == gscSkew3) {
					push @redundant, $query;
					unshift @pool, $pool[$i];
					redo POOL;
				}
			}
		}
		# No clustering happened with this guy
		push @unique, $query;
	}
	push @pool, @unique;
	$m->setFeature(\@pool);
	return \@redundant;
}

#----- Monitor (for debugging clustering) -----#
sub _monitor {
	my ($cmp) = @_;
	if ($cmp == gscIdentical)   {print "="}
	elsif ($cmp == gscOverlap)  {print "?"}
	elsif ($cmp == gscSuperset) {print "+"}
	elsif ($cmp == gscSubset)   {print "-"}
	elsif ($cmp == gscLong5)    {print ">"}
	elsif ($cmp == gscLong3)    {print "<"}
	elsif ($cmp == gscShort5)   {print "["}
	elsif ($cmp == gscShort3)   {print "]"}
	elsif ($cmp == gscSkew5)    {print "\\"}
	elsif ($cmp == gscSkew3)    {print "\/"}
	else {print "_"}
}

#----------------------------------------------------------------------
# Join
#----------------------------------------------------------------------

=head2 Join()

Description: Creates a new Tool by joining two others. The two
Tools may be simple and/or complex. Join takes three arguments,
the name of the new Tool, and the names of the extant Tools to be
joined.

 Usage: $new_tool = Bio::GSC::Tool::Join($name, $m1, $m2);

=cut

sub Join {
	my ($id, $m1, $m2) = @_;
	my $tool = new Bio::GSC::Tool;

	my ($type1, $type2) =
		($m1->getType(), $m2->getType());
	my ($f1, $f2) =
		($m1->getFeature(), $m2->getFeature());

	if ($type1 eq $type2) {$tool->setType($type1)}
	else {$tool->setType("$type1+$type2")}

	#----- simply push them into the same list
	my $feature;
	push(@$feature, @$f1, @$f2);
	$feature = sortFeatures($feature, 'Begin');
	$tool->setFeature($feature);
	return $tool;
}

#----------------------------------------------------------------------
# Intersect
#----------------------------------------------------------------------

=head2 Intersect()

Description: Creates a new Tool using the intersection of two others.
There are three "modes" and three "operators" (see below). You must
also specify a  boundary for the comparison. This interger
value determines how close the features must be to each other. Using
the coorOperator, it is the difference between Begin and End attributes.
Using the seqOperator, it is the amount of extra sequence grabbed from
the genomic. Using the scoreOperator, it is the cutoff score for the
comparison to be "true".

 Mode                    Behavior
 ---------------------   -----------------------------------------------
 gscSimpleFeatures        complex features are treated as simple
 gscFlattenedFeatures     complex features are converted to simple
 gscIdenticalFeatures     all subfeatures must be similar
 gscIncompleteFeatures    contiguous subfeatues must be similar
 gscOverlappingFeatures   some subfeatures must be similar

 Operator        Behavior
 --------------  -------------------------------------------------
 gscCoordinates  the Begin and End of each feature is compared
 gscGenomic      the genomic sequence is used for the comparison
 gscScore        the sbjct (or genomic) is used for the comparison


 Usage: $new_tool = Bio::GSC::Tool::Intersect($tool1,
                      $tool2, MODE, OPERATOR, $boundary);

=cut

sub Intersect {
	my ($m1, $m2, $mode, $op, $bound) = @_;
	my $tool = new;
	my ($f1, $f2);
	if ($mode == gscFlattenedFeatures) {
		$f1 = $m1->getFlatFeature;
		$f2 = $m2->getFlatFeature;
	}
	else {
		$f1 = $m1->getFeature;
		$f2 = $m2->getFeature;
	}

	# make sure the tool have features
	return $tool unless defined @$f1;
	return $tool unless defined @$f2;

	# keep only those features that are the "same"
	my @keep;
	my @unique;
	my @query = @$f1;
	my @sbjct = @$f2;

	POOL: while (@query) {
		my $query = shift @query;

		SEARCH: foreach my $sbjct (@sbjct) {
			my $cmp = $query->cmpFeature($sbjct, $mode, $op, $bound);
			next SEARCH if $cmp == FALSE;

			# Identical a match in any mode
			if ($cmp == gscIdentical) {
				push @keep, $query;
				next POOL;
			}

			# Any non-FALSE is TRUE in overlap mode
			if ($mode == gscOverlappingFeatures) {
				push @keep, $query;
				next POOL;
			}

			# Take the longest incomplete feature
			if ($mode == gscIncompleteFeatures) {
				if ($cmp == gscSuperset or $cmp == gscLong5 or
						$cmp == gscLong3 or $cmp ==gscSkew5) {
					push @keep, $query;
					next POOL;
				}
				elsif ($cmp == gscSubset or $cmp == gscShort5 or
						$cmp == gscShort3 or $cmp == gscSkew3) {
					push @keep, $query;
					next POOL;
				}
			}
		}
		# No intersection with this query
		push @unique, $query;
	}
	$tool->setFeature(\@keep);
	_setTrash(\@unique);
	return $tool;
}

#----------------------------------------------------------------------
# Exclude
#----------------------------------------------------------------------

=head2 Exclude()

Description: Creates a new Tool excluding the intersection of the two
compared. See Intersect() for more information.

 Usage: $new_tool = Bio::GSC::Tool::Exclude($tool1,
                        $tool2, MODE, OPERATOR, $boundary);

=cut

sub Exclude {
	my ($m1, $m2, $mode, $op, $bound) = @_;
	my $tool = new;

	my ($f1, $f2);
	if ($mode == gscFlattenedFeatures) {
		$f1 = $m1->getFlatFeature;
		$f2 = $m2->getFlatFeature;
	}
	else {
		$f1 = $m1->getFeature;
		$f2 = $m2->getFeature;
	}

	# make sure the tools have features
	return $tool unless defined @$f1;   # a given
	return $m1 unless defined @$f2;     # nothing to compare

	# keep only those features that are different
	my @unique;
	my @keep;
	my @query = @$f1;
	my @sbjct = @$f2;

	POOL: while (@query) {
		my $query = shift @query;

		SEARCH: foreach my $sbjct (@sbjct) {
			my $cmp = $query->cmpFeature($sbjct, $mode, $op, $bound);
			next SEARCH if $cmp == FALSE;

			# Identical a match in any mode
			if ($cmp == gscIdentical) {
				push @keep, $query;
				next POOL;
			}

			# Any non-FALSE is TRUE in overlap mode
			if ($mode == gscOverlappingFeatures) {
				push @keep, $query;
				next POOL;
			}

			# Take the longest incomplete feature
			if ($mode == gscIncompleteFeatures) {
				if ($cmp == gscSuperset or $cmp == gscLong5 or
						$cmp == gscLong3 or $cmp ==gscSkew5) {
					push @keep, $query;
					next POOL;
				}
				elsif ($cmp == gscSubset or $cmp == gscShort5 or
						$cmp == gscShort3 or $cmp == gscSkew3) {
					push @keep, $query;
					next POOL;
				}
			}
		}
		# No intersection with this query - so keep it
		push @unique, $query;
	}
	$tool->setFeature(\@unique);
	_setTrash(\@keep);
	return $tool;
}

#----------------------------------------------------------------------
# _Trash
#----------------------------------------------------------------------

=head2 _getTrash()

Description: Any time you call Intersect or Exclude, there is a list
of features that are thrown out. The last list can be retrieved with
this method.

 Usage: $list_ref = Bio::GSC::Tool::_getTrash();

=cut

sub _getTrash {
	return $Trash;
}
sub _setTrash {
	my ($Trash) = @_;
}


#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

