package Bio::GSC::Tool::Blast;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Tool::Blast

=head1 DESCRIPTION

Bio::GSC::Tool::Blast is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Tool::Blast is a plug-in parser for WU-BLAST (W. Gish). It
will not work with other BLAST algorithms. Please do not send requests
for parsers for other types of BLAST. Feel free to modify this parser
to make your own, non-WU-BLAST plugin. Blast features are complex, that
is, features contain subfeatures. Each "Sbjct" in the report may contain
mulitple high scoring pairs (HSPs). Each HSP is a feature, while each
Sbjct is also a feature. See Bio::GSC::Feature for more information.

=head1 DEPENDENCIES

 Bio::GSC::Feature

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use Bio::GSC::Feature;
use Bio::GSC::Constants;
use Bio::GSC::Utilities('getGap');

#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------
my $SeqON = FALSE;
sub ParseSbjctSequences {
	($SeqON) = @_;
}

#----------------------------------------------------------------------
# parse
#----------------------------------------------------------------------
sub parse {
	my ($this, $file, @filter) = @_;

	#----- some regular expressions used frequently in parsing
	my $Int = '\d+';
	my $Float = '[\de\.\-\+]+';
	my $Seq = '[A-Z\-\*\+\|]+';

	open(BLAST, "<$file") or die "ERROR: can't open $file\n";
	while(<BLAST>) {last if $_ =~ /^>/} # strip off header
	my $line = $_;  # $line is used somewhat globally

	SBJCT: {
		last SBJCT if blastDone($line);
		my $sbjct = new Bio::GSC::Feature;
		my $def = $line;
		DEF_LINE: {
			$line = <BLAST>;
			last SBJCT if blastDone($line);
			if ($line =~ /^\s+Length = \d+/) {last DEF_LINE}
			else {$def .= $line}
			redo DEF_LINE;
		}
		my ($len) = $line =~ /Length = (\d+)/;
		$def =~ s/\s+/ /g;
		my ($id) = $def =~ /^>\w*\|(\w+)/;
		$sbjct->setSbjctDef($def);
		$sbjct->setSbjctBegin(1);
		$sbjct->setSbjctEnd($len);
		$sbjct->setSbjctID($id);

		HSP: {
			CHECK: {
				last CHECK if $line =~ /^\s+Score/;
				last SBJCT if blastDone($line);
				$line = <BLAST>;
				redo CHECK;
			}
			my $hsp = new Bio::GSC::Feature;
			my ($score) = $line =~ /Score = ($Int)/;
			my ($p)     = $line =~ /[Sum ]*P[\(\d+\)]* = ($Float)/;
			$line = <BLAST>;
			my ($percent) = $line =~ /Identities = \d+\/\d+ \((\d+)%\)/;
			$hsp->setScore($score);
			$hsp->setP($p);
			$hsp->setPercent($percent);
			<BLAST>;

			SCORE: {
				my $first = 1;
				my ($q_begin, $q_seq, $q_end); # query begin, seq, end
				my $a_seq;                     # alignment between
				my ($s_begin, $s_seq, $s_end); # sbjct begin, seq, end
				$q_seq = $a_seq = $s_seq = '';
				my ($ql, $al, $sl);            # each line

				ALIGNMENT: {
					if ($first) {$ql = <BLAST>}
					else        {$ql = $line} 
					$al = <BLAST>;
					$sl = <BLAST>;

					my ($qb, $qs, $qe) = $ql =~ /($Int) ($Seq) ($Int)/;
					my $offset = index($ql, $qs);
					my $as = substr($al, $offset); chomp $as;
					my ($sb, $ss, $se) = $sl =~ /($Int) ($Seq) ($Int)/;
					($q_begin, $s_begin) = ($qb, $sb) if $first;
					($q_end, $s_end) = ($qe, $se);
					$q_seq .= $qs;
					$a_seq .= $as;
					$s_seq .= $ss;
					$first = 0;
					<BLAST>;
					$line = <BLAST>;
					redo ALIGNMENT if $line =~ /^Query/;
				}

				#----- GAPS -----
				my $q_gap = getGap(\$q_seq);
				my $s_gap = getGap(\$s_seq);
				$hsp->setQueryGap($q_gap);
				$hsp->setSbjctGap($s_gap);

				#----- Sequences ON? -----
				if ($SeqON) {
					$hsp->setQuerySeq($q_seq);
					$hsp->setSbjctSeq($s_seq);
					$hsp->setAlignment($a_seq);
				}

				#----- Save or filter -----
				$hsp->setBegin($q_begin);
				$hsp->setEnd($q_end);
				$hsp->setSbjctBegin($s_begin);
				$hsp->setSbjctEnd($s_end);
				$sbjct->addSubFeature($hsp)
					if ($hsp->queryFeature($filter[1]));

				redo HSP if $line =~ /^\s+Score|^\s+Minus Strand HSPs:/;

				#----- all HSP's might fall below filter -----
				unless ($sbjct->isComplex()) {
					$line = <BLAST>;
					redo SBJCT;
				}
				$this->addFeature($sbjct)
					if $sbjct->queryFeature($filter[0]);
				$line = <BLAST>;
				redo SBJCT;
			}
		}
	}
	close BLAST;
}

sub blastDone {
	my ($line) = @_;
	return 1 unless defined $line;
	if ($line =~ /^Parameters/) {return 1}
	elsif ($line =~ /^FATAL/) {return 1}
	else {return 0}
}

#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

 see other Bio::GSC modules.

=head1 AUTHOR

 Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

 http://genome.wustl.edu/gsc/staff/ikorf/Bio::GSC

=head1 ACKNOWLEDGEMENTS

 This software was developed at the Genome Sequencing Center at
 Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

 Copyright (C) 1998 Washington University, St. Louis, MO, USA.
 All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

