package Bio::GSC::Tool::Genscan;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Tool::Genscan

=head1 DESCRIPTION

Bio::GSC::Tool::Genscan is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Tool::Genscan is a plug-in parser for Genscan (Burge and
Karlin). Genscan features are a mixture of simple and complex features.
Each exon and suboptimal exon is simple. Each gene is a complex feature
containing exons. See Bio::GSC::Feature for more information.

=head1 DEPENDENCIES

 Bio::GSC::Feature

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use Bio::GSC::Feature;

#----------------------------------------------------------------------
# Class Variables/Methods
#----------------------------------------------------------------------


#----------------------------------------------------------------------
# parse
#----------------------------------------------------------------------

sub parse {
	my ($this, $file, @filter) = @_;
	open(GEN, "<$file") or die "ERROR: can't open $file\n";
	while(<GEN>) {last if $_ =~ /^-----/} # strip off header
	my $line = <GEN>; # dump off the blank line
	my $gene_id = 0;

	GENE: {
		$gene_id++;
		my $gene = new Bio::GSC::Feature;
		$gene->setSbjctID("genscan.$gene_id");
		$gene->setSbjctDef("Genscan predicted gene $gene_id");
		EXON: {
			$line = <GEN>;
			last GENE if $line =~ /^Predicted|^Suboptimal|^NO EXONS/;
			if ($line !~ /\w/) {
				#----- check for filtering
				redo GENE unless $gene->isComplex();
				$this->addFeature($gene)
					if $gene->queryFeature($filter[0]);
				redo GENE;
			}
			($line) = $line =~ /^\s*(\d.+)\s$/;
			my @field = split(/\s+/, $line);

			my $exon = new Bio::GSC::Feature;
			$exon->setSbjctID($field[0]);
			$exon->setName($field[1]);
			$exon->setBegin($field[3]);
			$exon->setEnd($field[4]);
			if     ($field[1] eq 'Sngl' or
					$field[1] eq 'Init' or
					$field[1] eq 'Intr' or
					$field[1] eq 'Term') {
				$exon->setP($field[11]);
				$exon->setScore($field[12]);
			}
			else {
				$exon->setScore($field[6]);
				redo EXON; #throwing out Prom and PlyA
			}

			$gene->addSubFeature($exon)
				if $exon->queryFeature($filter[1]);
			redo EXON;
		}
	}

	return unless $line =~ /^Suboptimal/;
	while(<GEN>) {last if $_ =~ /^-----/} # strip off header
	$line = <GEN>;
	SUBOPT: {
		$line = <GEN>;
		last SUBOPT if $line !~ /\w/;
		my @field = split(/\s+/, $line);
		my $exon = new Bio::GSC::Feature;
		$exon->setSbjctID($field[0]);
		$exon->setBegin($field[3]);
		$exon->setEnd($field[4]);
		$exon->setP($field[11]);
		$exon->setScore($field[12]);
		$this->addFeature($exon) if $exon->queryFeature($filter[1]);
		redo SUBOPT;
	}
	close GEN;
}


#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

 see other Bio::GSC modules.

=head1 AUTHOR

 Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

 http://genome.wustl.edu/gsc/staff/ikorf/Bio::GSC

=head1 ACKNOWLEDGEMENTS

 This software was developed at the Genome Sequencing Center at
 Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

 Copyright (C) 1998 Washington University, St. Louis, MO, USA.
 All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

