package Bio::GSC::Utilities;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Bio::GSC::Utilities

=head1 DESCRIPTION

Bio::GSC::Utilities is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Utilities contains several frequently-used functions that
may be useful outside the Bio::GSC scope.

=head1 ENVIRONMENT

 FASTA_EXTERNAL      name of the program for fetching fasta files
 GENBANK_EXTERNAL    ... genbank files
 EMBL_EXTERNAL       ... embl files

=head1 DEPENDENCIES

In order to fetch sequence information from a source other than a
flat file, it is necessary to have a program that prints flat files
to STDOUT. This can be an FTP client or a local database. The name
of the program must be stored in the appropriate environment variable.

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
	fcmp
	fetchFasta
	getFasta
	putFasta
	cloneObject
	getGap);

my $Fasta_External   = $ENV{FASTA_EXTERNAL};
my $Genbank_External = $ENV{GENBANK_EXTERNAL};
my $EMBL_External    = $ENV{EMBL_EXTERNAL};

#----------------------------------------------------------------------
# Class Variables
#----------------------------------------------------------------------

#----------------------------------------------------------------------
# fcmp
#----------------------------------------------------------------------

=head2 fcmp()

Description: Compares two number with fuzzy coordinates. Like most
comparison operators, it returns -1, 0, or 1 depending if the first
argument is smaller, equal, or greater than the second argument. fcmp
uses a fuzzy value that acts as a + or - that value for equality.

 Usage: fcmp($number1, $number2, $fuzzy);

=cut

sub fcmp {
	my ($x, $y, $fuzzy) = @_;
	if (abs($x - $y) < $fuzzy) {return 0}
	elsif ($x > $y) {return 1}
	else {return -1}
}

#----------------------------------------------------------------------
# fetchFasta
#----------------------------------------------------------------------

=head2 fetchFasta()

Description: Gets a fasta file using external methods. Returns the
definition line and the sequence.

 Usage: ($def, $seq) = getseq($gi_or_accession);

=cut

sub fetchFasta {
	my ($gi) = @_;
	open(GETSEQ,"$ENV{FASTA_EXTERNAL} $gi |") or return 0;
	my $def = <GETSEQ>;
	if (not defined $def or $def =~ /not found$/) {return 0}
	my @seq = <GETSEQ>;
	close GETSEQ;
	chomp(@seq);
	my $seq = join('', @seq);
	$seq = uc($seq);
	return($def, $seq);
}

#----------------------------------------------------------------------
# getFasta
#----------------------------------------------------------------------

=head2 getFasta()

Description: Gets a fasta file from the file system.

 Usage: ($def, $seq) = getFasta($filename);

=cut

sub getFasta {
	my ($file) = @_;
	open(FASTA,"<$file") or return 0;
	my $def = <FASTA>;
	chomp($def);
	my @seq = <FASTA>;
	close FASTA;
	chomp(@seq);
	my $seq = join('', @seq);
	$seq = uc($seq);
	return ($def, $seq);
}

#----------------------------------------------------------------------
# putFasta
#----------------------------------------------------------------------

=head2 putFasta()

Description: Writes a fasta file to the filesystem. You can specify the
number of characters per line (default is all on one line).

 Usage: putFasta($file, $def, $seq, $columns);

=cut

sub putFasta {
	my ($file,$def,$seq,$columns) = @_;
	open(FASTA,">$file") or return 0;
	print FASTA $def;
	if (defined $columns) {
		for(my $i=0;$i<length($seq)-1;$i+=$columns)
			{print FASTA substr($seq,$i,$columns), "\n"}
	}
	else {print FASTA $seq}
	close FASTA;
}

#----------------------------------------------------------------------
# postFasta
#----------------------------------------------------------------------

=head2 postFasta()

Description: Writes a fasta file to STDOUT. You can specify the
number of characters per line (default is all on one line).

 Usage: postFasta($def, $seq, $columns);

=cut

sub postFasta {
	my ($def,$seq,$columns) = @_;
	print $def;
	if (defined $columns) {
		for(my $i=0;$i<length($seq)-1;$i+=$columns)
			{print substr($seq,$i,$columns), "\n"}
	}
	else {print $seq}
}


#----------------------------------------------------------------------
# serialize
#----------------------------------------------------------------------

=head2 serialize()

Description: Serializes the object and hands back a refrence to a list
of output lines. An optional argument is data compression.

 Compression            Effect
 -------------------    --------------------------------------
 gscDataCompression0    no compression, very readable, default
 gscDataCompression1    not very readable, degenerate NT okay
 gscDataCompression2    not very readable, degenerate NT okay
 gscDataCompression3    not very readable, base-perfect sequence

 Usage: Bio::GSC::Utilities::serialize();

=cut

sub serialize {
	my ($data, $level) = @_;
    $level = 1 unless defined $level; 
    my ($i,$key,$value);
    my $tab = "  " x $level; 
    my $ref = ref($data);
    if (not $ref) { 
        if (defined $data) {print "$data\n"}
        else {print "UNDEFINED VALUE\n"}
    }
    elsif ($ref eq 'ARRAY') {
        print "$ref\n";
        $level++;
        for($i=0;$i<@$data;$i++) {
            print $tab;
            serialize($data->[$i],$level);
        }       
    }
    elsif ($ref eq 'HASH') {
        print "$ref\n";
        $level++;
        foreach $key (sort keys %$data) {
            print $tab,"$key = ";
            serialize($data->{$key},$level);
        }       
    }
    elsif ($ref =~ /^Bio::GSC/) {
        print "$ref\n";
        $level++;
        foreach $key (sort keys %$data) {
            print $tab,"$key = ";
            serialize($data->{$key},$level);
        }       
    }
    else {  
        print die "ERROR\n";
    }
}

#----------------------------------------------------------------------
# getGap
#----------------------------------------------------------------------

=head2 getGap()

Description: Takes a reference to a string and returns a reference to
a hash (key = position, value = length of gap).

 Usage: Bio::GSC::Utilities::getGap($seq_ref);

=cut

sub getGap {
	my ($seq) = @_;
	my %gap;
	my $offset = 0;     # current position in sequence
	my $gap_begin;      # position of first gap coor extending
	my $gap_total = 0;  # total number of gap characters seen so far
	FIND_GAP: {
		$gap_begin = index($$seq, "-", $offset);
		last FIND_GAP if $gap_begin == -1;
		$offset = $gap_begin;
		my $gap_length = 0; # length of the current gap
		EXTEND_GAP: {
			if (substr($$seq, $offset, 1) eq "-") {
				$offset++;
				$gap_length++;
				redo EXTEND_GAP;
			}
		}
		$gap{$gap_begin - $gap_total} = $gap_length;
		$gap_total += $gap_length;
		redo FIND_GAP;
	}
	return \%gap;
}
#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

