package Bio::GSC::Visual;
#----------------------------------------------------------------------
# General Documentation - Header
#----------------------------------------------------------------------

=head1 NAME Visual

=head1 SYNOPSIS

=head1 DESCRIPTION

Bio::GSC::Visual is part of the Bio::GSC bundle of perl modules
developed at the Genome Sequencing Center at Washington University,
St. Louis, MO, USA.

Bio::GSC::Visual contains methods for displaying sequence objects in a
graphic interchange format (GIF).

=head1 ENVIRONMENT

=head1 DEPENDENCIES

GD.pm

=head1 API

The application programming interface is described below. Class methods
begin with an uppercase letter, instance methods with a lowercase
letter, and private methods with an underscore (these may not even be
documented).

=cut

#----------------------------------------------------------------------
# Directives
#----------------------------------------------------------------------
require 5.004;
require Exporter;
use strict;
use vars qw ($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = "0.9";
@ISA = qw(Exporter);           
@EXPORT = qw();
@EXPORT_OK = qw();

use GD;
use PostData;

#----------------------------------------------------------------------
# Class Variables
#----------------------------------------------------------------------

my $TextArea = 100;
my $TopMargin = 20;
my $LeftMargin = 20;
my $RightMargin = 20;
my $BottomMargin = 20;
my $Size = 800;
my $CanvasSize = $Size - $LeftMargin - $RightMargin - $TextArea;
my $RowSize = 41;
my $Background = 'BLACK';
my $Foreground = 'WHITE';

sub inverse {
	($Background, $Foreground) = ($Foreground, $Background);
}

sub setSize {
	($Size) = @_;
	$CanvasSize = $Size - $LeftMargin - $RightMargin - $TextArea;
}

#----------------------------------------------------------------------
# Method Documentation
#----------------------------------------------------------------------

=head2 displayMarkup()

Description: creates a GIF image.

 Usage: $seq_object->display()

=cut

sub displayMarkup {
	my ($this, @met) = @_; # met is a list of tools and metrics

	#----- get the scale of the sequence
	my $scale = $CanvasSize / $this->getLength();
	my $length = $this->getLength();

	#----- get the size of the feature table
	my $rowsize = (@met + 1) * $RowSize + $TopMargin + $BottomMargin;

	#----- set up the image
	my $image = new GD::Image($Size, $rowsize);
	my %Color = setColorTable($image);
	$image->interlaced('true');
	$image->filledRectangle(0, 0, $Size, $rowsize, $Color{$Background});

	#----- display some header stuff
	my $intro = "Bio::GSC Sequence Display version 0.01";
	$image->string(
		gdLargeFont,
		$LeftMargin,
		$TopMargin / 2, 
		$intro,
		$Color{$Foreground});

	#----- display genomic information
	$image->string(
		gdMediumBoldFont,
		$LeftMargin,
		$TopMargin + $RowSize / 2,
		$this->getDefinition(),
		$Color{$Foreground});

	for(my $i=0;$i<@met;$i++) {
		my $m     = $met[$i];
		my $type  = $m->{NAME};
		my $color = $Color{$m->{COLOR}};
		my $high  = exists $m->{HIGH}  ? $m->{HIGH}  : 1;
		my $low   = exists $m->{LOW}   ? $m->{LOW}   : 0;
		my $att   = exists $m->{SCORE} ? $m->{SCORE} : 'Score';

		plotName($image, $i + 1, $type, $scale, $Color{$Foreground});
		if (exists $this->{'Tool'}{$type}) {
			my $tool = $this->getTool($type);
			plotTool($image, $tool, $i + 1, $scale, $color,
				$high, $low, $att);
		}
		elsif (exists $this->{'metric'}{$type}) {
			my $metric = $this->getMetric($type);
			plotMetric($image, $metric, $i + 1, $scale, $color,
				$high, $low, $att);
		}
	}

	#----- return the gif
	return $image->gif;
}

sub setColorTable {
	my ($image) = @_;
	my %Color = (
		BLACK     => $image->colorAllocate(  0,   0,   0),
		GRAY75    => $image->colorAllocate( 64,  64,  64),
		GRAY50    => $image->colorAllocate(128, 128, 128),
		GRAY25    => $image->colorAllocate(192, 192, 192),
		WHITE     => $image->colorAllocate(255, 255, 255),
		RED       => $image->colorAllocate(255,   0,   0),
		DARKRED   => $image->colorAllocate(128,   0,   0),
		ORANGE    => $image->colorAllocate(255,  96,   0),
		AMBER     => $image->colorAllocate(255, 192,   0),
		YELLOW    => $image->colorAllocate(255, 255,   0),
		GREEN     => $image->colorAllocate(  0, 255,   0),
		DARKGREEN => $image->colorAllocate(  0, 128,   0),
		CYAN      => $image->colorAllocate(  0, 255, 255),
		BLUE      => $image->colorAllocate(  0,   0, 255),
		DARKBLUE  => $image->colorAllocate(  0,   0, 128),
		VIOLET    => $image->colorAllocate(192,   0, 255),
		MAGENTA   => $image->colorAllocate(255,   0, 255),
	);
	return %Color;
}

sub plotName {
	my ($image, $row, $name, $scale, $color) = @_;
	$image->string(
		gdMediumBoldFont,
		$CanvasSize + $LeftMargin + 10,
		$row * $RowSize + $RowSize / 1.25, 
		$name,
		$color);
}

sub plotTool {
	my ($image, $tool, $row, $scale, $color, $high, $low, $att) = @_;

	#----- draw the plus/minus strand separator
	$image->line(
		$LeftMargin,
		$TopMargin + $row * $RowSize + $RowSize/2 -1,
		$LeftMargin + $CanvasSize,
		$TopMargin + $row * $RowSize + $RowSize/2 -1,
		$color);

	#----- draw the feature boxes on the appropriate strand
	my $featurelist = $tool->getFlatFeature();
	return unless defined @$featurelist and @$featurelist > 0;

	for(my $i=0;$i<@$featurelist;$i++) {
		my $f = $featurelist->[$i];
		my $begin = int $f->getBegin() * $scale;
		my $end = int $f->getEnd() * $scale;
		my $strand = $f->getStrand();
		my $pm; # holds the plus/minus value

		if ($strand > 0) {
			$pm = 1;
		}
		elsif ($strand < 0) {
			$pm = $RowSize / 2 - 1;
			($begin, $end) = ($end, $begin);
		}
		my $y = $TopMargin + ($row * $RowSize);
		$image->filledRectangle(
			$LeftMargin + $begin,        # x1 
			$y + $pm,                    # y1
			$LeftMargin + $end,          # x2
			$y + $pm + $RowSize / 2 - 2, # y2
			$color);
	}
}

sub plotMetric {
	my ($image, $metric, $row, $scale, $color, $high, $low, $att) = @_;

	#----- draw the plus/minus strand separator
	$image->line(
		$LeftMargin,
		$TopMargin + $row * $RowSize + $RowSize/2 -1,
		$LeftMargin + $CanvasSize,
		$TopMargin + $row * $RowSize + $RowSize/2 -1,
		$color);

	#----- assume normalization to 1.0 for now (Xpound)
	my $lastx = $LeftMargin;
	my $lasty = $TopMargin + $row * $RowSize + $RowSize/2 - 1;
	for(my $i=0;$i<@{$metric->{'plus'}};$i++) {
		my $thisx = $LeftMargin + $i * $scale;
		my $thisy = $TopMargin - ($metric->{'plus'}[$i] *
			$RowSize / 2) + ($row * $RowSize) + $RowSize/2 - 1;
		$image->line($lastx, $lasty, $thisx, $thisy, $color);
		($lastx, $lasty) = ($thisx, $thisy);
	}

	$lastx = $LeftMargin;
	$lasty = $TopMargin + $row * $RowSize + $RowSize/2 - 1;
	for(my $i=0;$i<@{$metric->{'minus'}};$i++) {
		my $thisx = $LeftMargin + $i * $scale;
		my $thisy = $TopMargin + ($metric->{'minus'}[$i] *
			$RowSize / 2) + ($row * $RowSize) + $RowSize / 2 - 1;
		$image->line($lastx, $lasty, $thisx, $thisy, $color);
		($lastx, $lasty) = ($thisx, $thisy);
	}
}


#----------------------------------------------------------------------
# General Documentation - Footer
#----------------------------------------------------------------------

=head1 SEE ALSO

see other Bio::GSC modules.

=head1 AUTHOR

Ian Korf (ikorf@sapiens.wustl.edu)

=head1 REFERENCES

http://sapiens.wustl.edu/~ikorf/BioGSC

=head1 ACKNOWLEDGEMENTS

This software was developed at the Genome Sequencing Center at
Washington Univeristy, St. Louis, MO.

=head1 COPYRIGHT

Copyright (C) 1998 Washington University, St. Louis, MO, USA.
All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind.

=cut


1;

