#!/usr/local/bin/perl

package Clink;
use strict;
use Cwd;
use File::Copy;
use Ace 1.50;
use Sequence;

sub new {

    my ($pkg,
	$name,  # Name of the clone
	$length # Must have a length
	) = @_;
    unless ( $length ) {
	die "Fatal Error: No DNA length data for: $name\n";
    }
    
    # Return new Clink object
    bless {
	_name => $name,     # The name of the clone
	_length => $length, # It's length in bp
	_oRight => [],      # Hash of format $clone{ position }
	                    # from Overlap_right positions
	_oLeft => [],       # As previous, from Overlap_left.
	_matches => {} # CrossMatch objects hashed on name of other Clink
	}, $pkg;
}

# Accessor methods

sub name { # Returns name; is read-only
    my $clink = shift;
    $clink->{'_name'};
}
sub length { # Returns length; is read-only
    my $clink = shift;
    $clink->{'_length'};
}
sub oRight {
    &_addCloneEnds( '_oRight', @_ );
}
sub oLeft {
    &_addCloneEnds( '_oLeft', @_ );
}
sub _addCloneEnds {
    my $var = shift;
    my $clink = shift;
    # Add clone/pos array to clink object or
    # return array of clone/pos arrays
    if ( @_ ) {
	my %cp = @_;
	# Overwrite existing Clone_ends
	foreach my $end ( @{$clink->{ $var }} ) {
	    #warn "Ends: $end->[0] $end->[1]\n";
	    if ( $cp{ $end->[0] } ) {
		$end->[1] = $cp{ $end->[0] };
		# Remove entry from hash
		delete $cp{ $end->[0] };
	    }
	}
	# Add new Clone_ends
	foreach my $name ( keys %cp ) {
	    push( @{$clink->{ $var }}, [$name, $cp{ $name }] );
	}
    } else {
	# Return array sorted on positions
	if (wantarray) {
	    return sort { $a->[1] <=> $b->[1] } @{$clink->{ $var }};
	} else {
	    return scalar @{$clink->{ $var }};
	}
    }
}
sub mapLeft {
    &_addMap( '_mapLeft', @_ );
}
sub mapRight {
    &_addMap( '_mapRight', @_ );
}

sub _addMap {
    my $var = shift;
    my $clink = shift;

    if ( @_ ) {
	$clink->{$var} = shift;
    } else {
	# Report value
	unless ( $clink->{$var} ) {
	    # Initialize if absent
	    $clink->{$var} = '';
	} else {
	    $clink->{$var};
	}
    }    
}

sub match {
    my $clink = shift;

    if (ref($_[0])) {
	# Have been passed a match to save
	my $match = shift;
	my $self = $clink->name();
	my $aName = $match->aName;
	$clink->{'_matches'}{$self} = $match;
	# Add match to hash under name of other clone
	if ($aName eq $self) {
	    my $bName = $match->bName();
	    $clink->{'_matches'}{$bName} = $match;
	} else {
	    $clink->{'_matches'}{$aName} = $match;
	}
    } else {
	# Return the match who's name matches string
	my $name = shift;
	if (my $match = $clink->{'_matches'}{$name}) {
	    return $match;
	} else {
	    return;
	}
    }
}

sub fromAceFile {
    my ($pkg,
	$file) = @_;
    my ($clink, @clinks);
    local $/ = ''; # Read in chunks separated by newlines
    open FILE, "< $file" or die
	"Fatal Error: Can't open file: $file: $!\n";
    while (<FILE>) {
	# Construct a clink object from each entry
	if ( /^Sequence\s+([:\w]+)\s*(\d+)?/m ) {
	    unless ( $2 ) {
		die "Fatal Error: No DNA length data for $1\n";
	    }
	    $clink = new Clink ( $1, $2 );
	} else {
	    next;
	}
	while ( /^Overlap_right\s+([:\w+]+)\s*(\d+)?/mg ) {
	    $clink->oRight( $1, $2 );
	}
	while ( /^Overlap_left\s+([:\w+]+)\s*(\d+)?/mg ) {
	    $clink->oLeft( $1, $2 );
	}
	push @clinks, $clink
    }
    close FILE;
    return @clinks; # Could return ref instead
}

sub toAce {
    my $clink = shift;
    my (@acelines);

    push (@acelines,
	  "\nSequence ", $clink->name(), " ", $clink->length() );
    foreach my $oRarray ( $clink->oRight() ) {
	push (@acelines, "\nOverlap_right @$oRarray");
    }
    foreach my $oRarray ( $clink->oLeft() ) {
	push (@acelines, "\nOverlap_left");
    }
    return join '', @acelines, "\n";
}

sub readAceData {
    my ($host,
	$port,
	$list,     # Ref to list of clones to fetch,
	           # or scalar containing name of a
	           # single clone, or '*' for all clones.
	$readEnds  # If def, then fetches Clone_end info
	) = @_;
    my ($db,
	@gen_seqs,
	@retList   # Array of clink objects to return
	);

    return unless $list;
    
    $db = slowConnect( $host, $port );

    # Might want to check if sequence is Finished?
    # -- No, I assume all data in humace is finished
    my $clone;
    if ( ref($list) ) {
	foreach my $name (@$list) {
	    if (ref($name) eq 'Ace::Object') {
		$clone = $name;
	    } else {
		$clone = $db->fetch('Genome_sequence', $name);
	    }
	    &_getClink( $clone, \@retList, $readEnds ) or
		warn " *** Couldn't find $name in database *** \n";
	}
    } else {
	if ( $list eq '*' ) {
	    my $i = $db->fetch_many('Genome_sequence','*');
	    while ($clone = $i->next) {
		&_getClink( $clone, \@retList, $readEnds );
	    }
	} else {
	    $clone = $db->fetch('Genome_sequence', $list) and
		&_getClink( $clone, \@retList, $readEnds );
	}
    }
    return @retList;
}

sub _getClink {
    my ($clone,
	$retRef,
	$readEnds) = @_;

    # Store length of DNA object.
    # Had big memory leak here, when this line was:
    # my $length = $clone->at('DNA[2]') or
    # because this doesn't dereference the AcePerl
    # object.
    my ($length) = $clone->at('DNA[2]') or
	warn "No Length: $clone\n" and return;
    
    # Make new Clink - $clone must be in qutoes,
    # or will pass whole AcePerl object to clink->new
    my $clink = new Clink( "$clone", $length );
    #print "$clone $length\n";

    if ($readEnds) {
	
	# Store info about clones with
	# left ends inside $clone
	if ( my @entries = $clone->at('Structure.Overlap_right[1]') ) {
	    foreach my $row ( @entries ) {
		my @row = $row->row();
		next if $row[0] eq $clone; # Not interested in info about self
		$clink->oRight( @row );
	    }
	}
	
	# Then do the same for right
	if ( my @entries = $clone->at('Structure.Overlap_left[1]') ) {
	    foreach my $row ( @entries ) {
		my @row = $row->row();
		next if $row[0] eq $clone; # Not interested in info about self
		$clink->oLeft( @row );
	    }
	}
    }
    
    # Add the new Clink to the array
    push @$retRef, $clink;
}

sub nowDeprecated_findEndMatches {
    my ($seqA,
	$seqB) = @_;

    # This was the plan, but cross_match is so fast,
    # that it isn't necessary:
    # # Will look for $seqX.seq and $seqX.seq_500L +
    # # $seqX.seq_500R for both A and B in $seqdir.
    # # Does end matching if _500 files are found,
    # # which is faster.  Returns undef if match is
    # # not made (if one of the sequences needs
    # # complementing for example), or if no ends
    # # are found.  If sucessful, returns following
    # # variables: 

    my ($A_Bl, # Position of Left end of B in A
	$A_Br, # Position of Right end of B in A
	$B_Al, # Position of Left end of A in B
	$B_Ar, # Position of Right end of A in B
	$strand  # Strand which match is on - w=fwd
                 #                          - c=rev
	);
    
    my (@results);
    unless ( -r "$SEQ_CACHE/$seqA.seq" and -r "$SEQ_CACHE/$seqB.seq" ) {
	die "Fatal Error: $SEQ_CACHE/$seqA.seq, $seqB.seq not readable: $!\n";
    } else {
	open CROSS_MATCH,
	"cross_match -minmatch $MIN_MATCH -minscore 30 $SEQ_CACHE/$seqA.seq $SEQ_CACHE/$seqB.seq 2>/dev/null |"
	    or die "Fatal Error: Can't run cross_match: $!\n";
	while (<CROSS_MATCH>) {
	    next unless /\(\d+\).*\(\d+\)/;
	    tr/()//d;
	    push @results, $_;
	}
	close CROSS_MATCH;
	unlink "$SEQ_CACHE/$seqA.seq.log";
    }

    if ( scalar @results > 1 ) {
	# Only intersted in 1st and last lines of results
	# because output lines are sorted by position,
	# and we are only interested in the ends.
	@results = @results[0,$#results];
    }

    # Find end matches on forward strand
    foreach ( @results ) {
	unless ( / C / ) {
	    my ($score, $Psub, $Pdel, $Pins,
		$A, $A_beg, $A_fin, $A_remain,
		$B, $B_beg, $B_fin, $B_remain ) = split;
	    # Can only build links if have neither insertions nor deletions
	    if ( $Pdel == 0 and $Pins == 0 ) {
		# There are four possible ends to be found if
		# the sequences are linkable:
		$A_Bl = $A_beg if $B_beg == 1;
		$A_Br = $A_fin if $B_remain == 0;
		$B_Al = $B_beg if $A_beg == 1;
		$B_Ar = $B_fin if $A_remain == 0;
	    }
	}
    }

    # Look for end matches on opposite strand if
    #  we didn't get any on forward
    unless ( $A_Bl or $A_Br or $B_Al or $B_Ar ) {
	foreach ( @results ) {
	    if ( / C / ) {
		my ($score, $Psub, $Pdel, $Pins,
		    $A, $A_beg, $A_fin, $A_remain,
		    $Comp,
		    $B, $B_beg, $B_fin, $B_remain ) = split;
		# Can only build links if have neither insertions nor deletions
		if ( $Pdel == 0 and $Pins == 0 ) {
		    # There are four possible ends to be found if
		    # the sequences are linkable:
		    $A_Bl = $A_beg if $B_beg == 1;
		    $A_Br = $A_fin if $B_remain == 0;
		    $B_Al = $B_beg if $A_beg == 1;
		    $B_Ar = $B_fin if $A_remain == 0;
		}
	    }
	}
	# Did we get reverse hits?
	if ( $A_Bl or $A_Br or $B_Al or $B_Ar ) {
	    $strand = 'c'; # Crick
	}
    } else {
	# Hits were found on forward strand
	$strand = 'w'; # Watson
    }
    
    return ($strand, $A_Bl, $A_Br, $B_Al, $B_Ar);
}

#      sub   del   ins
#        .     .     .             .    
#  0     1     2     3      4      5      6      7      8      9     10     11
# 98  0.00  0.00  0.00  130N4      1    104  84065  92M18  68800  68903      0
# 98  0.00  0.00  0.00  92M18  68800  68903      0  130N4      1    104  84065
# 82510.00 0.00 0.00  130N4        1 84169 0    130N4        1 84169 0

#  103  0.00 0.00 0.00  CFAT5    20771 20874 (0)    A1280    1   104   (22149)
#  103  0.00 0.00 0.00  CFAT5    20771 20874 (0)  C A1280.RC (0) 22253 22150
#  103  0.00 0.00 0.00  CFAT5.RC 1   104 (20770)  C A1280   (22149)   104     1



# sub slowConnect {
#     my($host,
#        $port) = @_;
#     my $db;
#   ACECONNECT: {
#       my $tries;
#       until ($db) {
# 	  $db = Ace->connect(-host => $host,
# 			     -port => $port)
# 	      and last ACECONNECT;
# 	  $tries++;
# 	  sleep 30;
# 	  if ($tries > 39) { # More than 20 min of trying
# 	      die
# 		  "Error: Couldn't connect to $host on $port\n",
# 		  "       after $tries retries:", Ace->error(), "\n";
# 	  }
#       }
#   }
#     return $db;
# }


#------------------------------------------------------------

# The contig package deals with contigs built from
# map information
package Contig;
use CrossMatch;
use vars qw( @ISA );
@ISA = qw( CrossMatch );

sub new {
    my ($pkg, $map, $chr, @clinks) = @_;

    my $matcher = CrossMatch::Factory->new();

    $matcher->{'_map'} = $map; # The name of the map
    $matcher->{'_chr'} = $chr; # The chromosome it is on
    $matcher->{'_clones'} = [ @clinks ]; # Array of clink objects

    # Return a new Contig object
    return bless $matcher, $pkg;
}

# Accessor methods
sub name {
    my $contig = shift;
    @_ ? $contig->{'_map'} = shift : $contig->{'_map'};
}
sub chr {
    my $contig = shift;
    @_ ? $contig->{'_chr'} = shift : $contig->{'_chr'};
}
sub clones {
    my $contig = shift;
    @_ ? push( @{$contig->{'_clones'}}, @_ )
	: @{$contig->{'_clones'}};
}
sub nextClone {
    my $contig = shift;
    my $clink = shift @{$contig->{'_clones'}};
}
sub pullClone {
    my ($contig, $cloneName) = @_;
    foreach my $i ( 0..$#{$contig->{'_clones'}} ) {
	my $clink = $contig->{'_clones'}[$i];
        next unless $clink->name() eq $cloneName;
        splice( @{$contig->{'_clones'}}, $i, 1 );
        return $clink;
    }
}
sub clinkByName {
    my ($contig, $cloneName) = @_;
    foreach my $clink ( $contig->clones() ) {
	return $clink if $clink->name() eq $cloneName;
    }
    return;
}

sub sortByMap {
    my $contig = shift;
    # Sorts on left map position, or right, or name
    
    @{ $contig->{'_clones'} }
    = sort { $a->mapLeft() <=> $b->mapLeft() or
	     $a->mapRight() <=> $b->mapRight() or
	     $a->name() cmp $b->name() } @{$contig->{'_clones'}};
    
}

sub fetchDNAfromAce {
    my $contig = shift;
    my $db = shift;
    my( @DNA, @dir, $cacheDir );
    
    # Make a list of Clinks we can't find a
    # sequence file for
    foreach my $clink ($contig->clones()) {
	my $name = $clink->name();
	unless ($contig->findFile( $name )) {
	    push( @DNA, $name );
	}
    }
    return 1 unless scalar @DNA; # Found files for all seqs

    @dir = $contig->dir();
    foreach my $dir ( @dir ) {
	if (-w $dir) {
	    $cacheDir = $dir;
	    last;
	}
    }
    unless ($cacheDir) {
	croak "Can't find writable directory in: @dir";
    }

    foreach my $sequence (@DNA) {
	my( $seqObj, $dnaTag, $DNA, $len, $dnaObj );
	if (
	    # Fetch sequence
	    $seqObj = $db->fetch( 'Sequence', $sequence ) and

	    # Is there a DNA_contig?
	    $dnaTag = $seqObj->at('DNA[1]') and

	    # Get the DNA itself
	    ($DNA) = $dnaTag->fetch()->at()
	    ) {

	    # Sequence must contain at least one character
	    if (length($DNA)) {
		$fasta = Sequence->new_from_strings( $name, $DNA );
		# Check that we got valid dna characters only
		unless ($seqObj->isdna_iub) {
		    warn "Got bad DNA from acedb: $sequence\n";
		    next;
		}
	    } else {
		warn "Empty DNA object: $sequence\n$DNA";
		next;
	    }
	    $fasta->write_fasta_file( "$cacheDir/$sequence.seq" ); # Croaks if fails
	    
	} else {
	    # Want to know about any Sequences
	    # with missing DNA
	    warn "Can't fetch DNA for $sequence\n";
	}
    }
}

sub doMatches {
    my $contig = shift;
    my @contig = $contig->clones();
    
    # Delete any Clone_end info
    # -- don't need to do with new access functions
    foreach my $clink (@contig) {
	$clink->{'_oRight'} = [];
	$clink->{'_oLeft'} = [];
    }
    
    foreach my $i ( 0..$#contig ) {
	my $A = $contig[$i];
	my $Aname = $A->name();

	# Looks for match in previous 3 clones as well
	# if fails to find match in next
      SEARCH: foreach my $r ( reverse -3..-1, 1 ) {
	  my( $B, $Bname, $match );
	  my $n = $i + $r;
	  # Check $n is in range
	  next if ( ($n < 0) or ($n > $#contig) );
	  $B = $contig[$n];
	  $Bname = $B->name();

	  # End loop if already have a match between this pair
	  last if $A->match( $Bname );
	  
	  # Do cross_match, returning a new CrossMatch::Match object
	  $match = $contig->crossMatch( $Aname, $Bname );
	  # Filter object to hide non-end hits
	  $match->filter(\&CrossMatch::endHits);
	  # End search, saving match in both clinks if we got a hit
	  if ($match->count) {
	      $A->match( $match );
	      $B->match( $match );
	      last SEARCH;
	  }

# 	  my ($strand, $A_Bl, $A_Br, $B_Al, $B_Ar) = &Clink::findEndMatches( $Aname, $Bname );
# 	  if ( $strand eq 'w' ) { # Watson
# 	      $A->oRight( $Bname, $A_Bl ) if $A_Bl;
# 	      $A->oLeft( $Bname, $A_Br ) if $A_Br;
# 	      $B->oRight( $Aname, $B_Al ) if $B_Al;
# 	      $B->oLeft( $Aname, $B_Ar ) if $B_Ar;
# 	      last SEARCH; # Have found match
# 	  } elsif ( $strand eq 'c' ) { # Crick
# 	      if ($revList) {
# 		  $rev{ $i } = $Aname;
# 		  $rev{ $n } = $Bname;
# 	      } else {
# 		  #warn " *** $Aname matches complement of $Bname *** \n";
# 		  warn "Complement match: $Aname $Bname\n";
# 		  last SEARCH; # Have found match
# 	      }
# 	  }
      }
# 	unless ( $A->oRight() ) {
# 	    warn " *** Break in contig ", $contig->name(), " after $Aname *** \n";
# 	}
    }
}

sub toLinks {
    my $contig = shift;
    my $filter = shift;
    my @linksList;
    my $link;
    
  CLONE: while ( my $clink = $contig->nextClone() ) {
      # Create a new Link object unless one
      # already exists
      unless ($link) {
	  $link = Link->newFromClink( $clink );
	  $link->chr( $contig->chr() );
      }
      
      # Terminate Clink and add to return array
      # if it doesn't have any Clone_lefts
      unless ( $clink->oRight() ) {
	  # Ignore 1 subsequence links -- they're not links!
	  if ( $link->subSeq() > 1 ) {
	      # Add the link object to the return array
	      push @linksList, $link;
	  }
	  undef $link;
	  next;
      }
      
      # Look through clink object for all
      # clone left ends
      foreach my $left ( $clink->oRight() ) {
	  # Get name and position from the array
	  my ($rName, $rPos) = @$left;
	  my $rClink;

	  # Make $rPos an absolute postion
	  --$rPos;
	  
	  # Go onto the next if already in Link
	  next if $link->subSeq( $rName );
	  
	  # Get the Clink of correct name from the array
	  unless ($rClink = $contig->clinkByName( $rName )) {
	      # This can only happen if a number of clones
	      # are in the wrong orientation
	      warn "Error: ", $clink->name(), " and $rName: wrong orientation in ",
	      $contig->name(), " \n";
	      # Not interested in link!
	      undef $link;
	      # Go to next clone
	      next CLONE;
	  }
	      
	  # Add clink object to Link
	  $link->clones( $rClink );
	  
	  # Update the subSeq table
	  $link->addSubSeq( $rClink->name(), $clink->name(), $rPos, $rClink->length() );
      }
      
  }
    return @linksList;
}

sub toAce {
    my $contig = shift;

    my $map = $contig->name();
    my $chr = $contig->chr();
    my @acelines;

    @acelines = (
		 "\n//----------------------",
		 "\n\nMap $map",
		 "\nChromosome $chr"
		 );

    foreach my $clone ( $contig->clones() ) {
	push (@acelines,
	      "\n\nSequence ", $clone->name(), " ", $clone->length() );
	foreach my $oRarray ( $clone->oRight() ) {
	    push (@acelines, "\nOverlap_right @$oRarray");
	}
	foreach my $oRarray ( $clone->oLeft() ) {
	    push (@acelines, "\nOverlap_left");
	}
	if ( my $mapL = $clone->mapLeft() ) {
	    push (@acelines, "\nMap $map Ends Left $mapL" );
	}
	if ( my $mapR = $clone->mapRight() ) {
	    push (@acelines, "\nMap $map Ends Right $mapR" );
	}
	
    }
    return join '', @acelines, "\n";
}

#------------------------------------------------------------

# Links package builds link object from contig with
# correct Overlap_right and Overlap_left info
package Link;
use vars qw( @ISA );
@ISA = qw( clink );

sub new {
    my $pkg = shift;
    my $name = shift;
    #my $length = shift;

    unless ( $name ) {
	die "Fatal Error: Name not specified for new Link object\n";
    }

    # Return a new Link object
    bless {
	_name => $name, # The name of the link.
	_oldName => '', # For storing original name
                        # if it needs to be renamed.
	_clones => [],  # Array of clink objects
	_subSeqs => {}, # Hash of subSeq => [ start, stop ] objects.
	_chr => '',     # Chromosome it is on.
	_mapLink => ''  # Need to add code to do this properly.
	}, $pkg;
}

# Accessor methods
sub name {
    my $link = shift;

    if (@_) {
	# Must be changing name, since new Link object
	# always has a name

	# Record prefix if there
	my ($prefix) = $link->{'_name'} =~ /^(LINK_)/;

	my $newName = shift;
	$newName = $prefix . $newName;
	
	# Don't do anything if new name is same as old
	if ( $newName eq $link->{'_name'} ) {
	    return;
	} else {
	    # Mustn't rename twice
	    unless ( $link->{'_oldName'} ) {
		$link->{'_oldName'} = $link->{'_name'};
	    }
	    $link->{'_name'} = $newName;
	    return 1;
	}
    }
    return $link->{'_name'};
}
sub oldName {
    my $link = shift;
    return $link->{'_oldName'};
}
sub length {
    # Calculates the length of a link
}
sub chr {
    my $link = shift;
    @_ ? $link->{'_chr'} = shift : $link->{'_chr'};
}
sub clones {
    my $link = shift;
    @_ ? push @{ $link->{'_clones'} }, @_ : @{ $link->{'_clones'} };
}
sub subSeqHash {
    my $link = shift;
    return $link->{'_subSeqs'};
}
sub subSeq {
    my $link = shift;
    if ( @_ ) {
	my $name = shift;
	unless ( @_ ) {
	    if (defined $link->{'_subSeqs'}{ $name }) {
		return @{$link->{'_subSeqs'}{ $name }};
	    } else {
		return;
	    }
	} else {
	    my $start = shift;
	    my $stop = shift;
	    unless ( $stop ) {
		die "Fatal Error: Link: ", $link->name(),
		": $name Start- $start No stop position provided\n";
	    }
	    $link->{'_subSeqs'}{ $name } = [ $start, $stop ] and
		return 1;
	    # Successfully added position info
	}
    } else {
	# Returns sorted array of subseqs, or number
	# of subseqs in array context
	if ( wantarray ) {
	    return sort { $link->{'_subSeqs'}{$a}[0] <=> $link->{'_subSeqs'}{$b}[0] }
	    keys %{$link->{'_subSeqs'}};
	} else {
	    return scalar keys %{$link->{'_subSeqs'}};
	}
    }
}

sub newFromClink {
    my ($pkg,
	$clink) = @_;

    my $name = "LINK_" . $clink->name();
    my $link = Link->new( $name );

    $link->subSeq( $clink->name(), 1, $clink->length() );
    $link->clones( $clink );

    # Return a new Link object
    return $link;
}

sub addSubSeq {
    # General method for adding a new subsequence to
    # a link object
    my(
       $link, # Link object
       $new, # Name of new subSeq to add
       $parent, # Name of parent subSeq in link
       $pos, # ABSOLUTE coordinate of match in the parent
             # ie: first resudue of parent = 0
       $length # Length of subSeq
       ) = @_;

    # Is Link extended leftwards?
    if ( $pos < 0 ) {
	# Need to rename link
	$link->name( $new );

	# Offset to add to all subseqs
	my $offset = -1 * $pos;

	# Get list of subSeqs
	foreach my $sub ( $link->subSeq() ) {
	    # Get position of named subSeq
	    my ($x, $y) = $link->subSeq( $sub );
	    $x += $offset;
	    $y += $offset;
	    # Overwrite old positions with new
	    $link->subSeq( $sub, $x, $y );
	}
    }

    # Find the postion in the link of the clone
    # which the match is in:
    my ($pStart, $pStop) = $link->subSeq( $parent );

    # Calculate start and stop, and add info to Link
    my $start = $pStart + $pos;
    my $stop = $start + $length - 1;
    $link->subSeq( $new, $start, $stop );
}

sub getLink {
    my(
       $pkg,
       $linkName,
       $host,
       $port
       ) = @_;

    my( $db, $aceLink, $chr, $link, @clinkList );

    $db = SUPER::slowConnect( $host, $port );
    $aceLink = $db->fetch( 'Link', $linkName )
	or die "Fatal Error: No such Link: $linkName\n";
    # Make a new Link object
    $link = Link->new( $linkName );

    # Which Chromosme is it on?
    ($chr) = $aceLink->at('Origin.Chromosome[1]');
    $link->chr( $chr );

    # Get data for each subsequence
    foreach my $sub ( $aceLink->at('Structure.Subsequence[1]') ) {
	if ( my @su = $sub->row() ) {
	    # Add to subSeqs hash
	    $link->subSeq( @su );
	}
    }

    # Need to make sure Clinks are in correct
    # order in clones array.  Called in this 
    # way the subSeq method returns list of
    # postion sorted names
    foreach my $sub ( $link->subSeq() ) {
	# Is it a genome_sequence?
	if ( my $gen = $db->fetch('Genome_Sequence', $sub) ) {
	    # Add a clink object onto @clinkList; "1" reads Clone_end info
	    &Clink::_getClink( $gen, \@clinkList, 1 );
	}
    }
    # Add all the fetched Clinks to the array
    $link->clones( @clinkList );

    return $link;
}

sub addClinkToLink {
    my(
       $A,
       $link,
       $host,
       $port
       ) = @_;

    # Takes a Clink and Link object and tries to
    # add the Clink into the link.  Needs to know
    # an aceserver host and port to get the Link's
    # DNA for doing the cross_matching.  Returns
    # undef if fails, or a modified Link object,
    # which has properties to rename the old LINK_
    # entry correctly when toAce is called on it.

    # Returns 1 on success, 2 if the clink matches
    # in the reverse orientation, and undef on
    # failure

    my( @cloneList, $Aname, $fromAce);

    # _clones array must be sorted
    foreach my $clone ( $link->clones() ) {
	push @cloneList, $clone->name();
    }

    # Get the sequences for crossmatching
    &Clink::fetchDNA(\@cloneList,
		     $host,
		     $port
		     );
    
    $Aname = $A->name();
    
    # Try to find DNA for $clink
    # Is it in the cache?
    unless ( -f "$SUPER::SEQ_CACHE/$Aname.seq" ) {
	# Is it in the cwd?
	my $cwd = cwd();
	if ( -f "$cwd/$Aname.seq" ) {
	    # copy to SEQ_CACHE directory
	    copy( "$cwd/$Aname.seq", "$SUPER::SEQ_CACHE/$Aname.seq" );
	} else {
	    &Clink::fetchDNA([$Aname],
			     $host,
			     $port
			     );
	    # Need to record that we got it
	    # from acedb, because can't reverse-
	    # complement in acedb (yet).
	    $fromAce = 1;
	}
    }
    unless ( -f "$SUPER::SEQ_CACHE/$Aname.seq" ) {
	die "Fatal Error: Can't find DNA for $Aname\n";
    }

    
    foreach my $B ( $link->clones() ) {
	my $Bname = $B->name();
	
	my( $strand, $A_Bl, $A_Br, $B_Al, $B_Ar ) = &Clink::findEndMatches( $Aname, $Bname );
	if ( $strand eq 'w' ) { # Watson
	    $A->oRight( $Bname, $A_Bl ) if $A_Bl;
	    $A->oLeft( $Bname, $A_Br ) if $A_Br;
	    $B->oRight( $Aname, $B_Al ) if $B_Al;
	    $B->oLeft( $Aname, $B_Ar ) if $B_Ar;
	    
	    # Add to subseqs hash, and rename link and renumber
	    # if necessary
	    my( $rPos );
	    if ( $A_Bl ) {
		# Left end of B is in A
		$rPos = 1 - $A_Bl;
	    } elsif ( $B_Al ) {
		# Left end of A is in B
		$rPos = $B_Al - 1;
	    }
	    $link->addSubSeq( $Aname, $Bname, $rPos, $A->length() );

	    return 1; # Have found match
	} elsif ( $strand eq 'c' ) { # Crick
	    if ( $fromAce ) {
		warn
		    " *** $Aname matches complement of $Bname\n",
		    "     $Aname needs to be reverse-complemented in acedb *** \n";
	    } else {
		# And code here to rev-comp sequence, and redo
		warn
		    " *** $Aname matches complement of $Bname\n",
		    "     Sequence $Aname needs to be reverse-complemented *** \n";
	    }
	    return 2; # Have found match on reverse
	}
    }
    return;
}

sub toAce {
    my $link = shift;

    my @ace; # acedb format output

    # Commented spacer to make output easier to look through
    push (@ace,
	  "\n//--------------------\n");

    # Rename old entry if Link has grown leftwards
    if ( my $oldName = $link->oldName() ) {
	push (@ace,
	      "\n-R Link $oldName ", $link->name(), "\n");
    }

    # Format ace lines for new link object
    push (@ace,
	  "\nSequence ", $link->name(), "\n",
	  "Properties Link\n");
    # Chromosome name
    if ( my $chr = $link->chr() ) {
	#warn $link->name(), " added $chr\n";
	push (@ace, "Chromosome $chr\n");
    }
    # Subsequences
    my $subs = $link->subSeqHash();
    foreach my $sub(sort { $subs->{ $a }[0] <=> $subs->{ $b }[0] }
 		    keys %$subs ) {
 	push (@ace, "Subsequence $sub $subs->{ $sub }[0] $subs->{ $sub }[1]\n");
    }
    
    foreach my $clink ( $link->clones() ) {
	push (@ace,
	      "\nSequence ", $clink->name(), "\n",
	      "From Source ", $link->name() );
	foreach my $oRarray ( $clink->oRight() ) {
 	    push (@ace, "\nOverlap_right @$oRarray");
 	}
 	foreach my $oLarray ( $clink->oLeft() ) {
 	    push (@ace, "\nOverlap_left");
 	}
	push (@ace, "\n");
    }

    return join '', @ace, "\n";

}

#------------------------------------------------------------

return 1;

__END__

=head1 NAME Clink - manipulate acedb link objects

=head1 SYNOPSIS


=head1 DESCRIPTION

B<Clink.pm> facilitates the creation and management of acedb link
objects.  Link objects have B<Properties.Link>, and contain
B<Subsequence> objects which are genomic clones.

=head1 AUTHOR

B<James Gilbert> email jgrg@sanger.ac.uk

