
package CrossMatch::Factory;

require 5.004;

use strict;
use Carp;
use Cwd;

# Makes a new CrossMatch factory object
sub new {
    my $pkg = shift;
    my( @dir );

    # Default to cwd if no directories supplied
    if (@_) {
	@dir = _dirs(@_);
    } else {
	$dir[0] = cwd();
    }

    return bless {
	dir => [ @dir ],
	minMatch => '50',
	minScore => '50',
	_extn => ['']
	}, $pkg;
}

# Accessor methods for cross_match's minmatch and minscore parameters
sub minMatch {
    _changeInt( 'minMatch', @_ );
}
sub minScore {
    _changeInt( 'minScore', @_ );
}
sub _changeInt {
    my( $call, $matcher, $num ) = @_;

    if ($num) {
	$num =~ /^\d+$/
	    or croak qq(non-integer "$num" supplied to $call);
	$matcher->{$call} = $num;
    } else {
	return $matcher->{$call};
    }
}

sub dir {
    my $matcher = shift;

    if (@_) {
	$matcher->{'dir'} = [ _dirs(@_) ];
    } else {
	return @{$matcher->{'dir'}};
    }
}
sub _dirs {
    my @dirs = @_;
    foreach (@dirs) {
	# Strip trailing "/" from name
	s|/$||;
    }
    return @dirs;
}

sub extn {
    my $matcher = shift;
    my( @extn );

    if (@_) {
	foreach my $extn (@_) {
	    # Add a leading dot
	    $extn =~ s|^\.?|\.| unless $extn eq '';
	    push( @extn, $extn );
	}
	$matcher->{'_extn'} = [ @extn ];
    } else {
	# Return all extensions
	return @{$matcher->{'_extn'}};
    }
}

sub findFile {
    my $matcher = shift;
    my $name = shift;

    foreach my $dir ($matcher->dir()) {
	foreach my $file (map "$dir/$name$_", ($matcher->extn()) ) {
	    # Check for a readable file of this name
	    return $file if -r $file;
	}
    }
    # Return undef if we didn't find a matching filename
    return;
}

sub crossMatch {
    my $matcher = shift;
    my( @seq ) = (@_)[0,1];
    my( $minM, $minS );
    
    defined $seq[1]
	or croak qq(not enough arguments to crossMatch: "@_");
    
    # Find files with these names
    foreach (@seq) {
	my $name = $_;
	$_ = $matcher->findFile( $_ )
	    or croak qq(could not find a file for "$name");
    }
    

    ### Do cross_match on the found files ###
    $minM = $matcher->minMatch();
    $minS = $matcher->minScore();
    # Make new match object, saving parameters
    my $match = CrossMatch->new(@seq, $minM, $minS);
    open( CROSS_MATCH, "cross_match -minmatch $minM -minscore $minS @seq 2>/dev/null |" );
    while (<CROSS_MATCH>) {
	next unless /\(\d+\).*\(\d+\)/;
	tr/()//d;
	my @res = split;
	unless ($res[8] eq 'C') {
	    @res = (@res[0..7], 'W', @res[8..11]);
	}
	$match->hit( @res );
    }
    close CROSS_MATCH;
    unlink "$seq[0].log";
    if ($?) {
	croak "Error from cross_match: $!";
    }

    # Return Match object
    return $match;
}

package CrossMatch;

require 5.004;

use strict;
use Carp;
use Cwd;
use vars qw( $AUTOLOAD );

# Fields which record parametrs given to cross_match
my %param = (
	     aFile => 1,
	     bFile => 1,
	     minM  => 1,
	     minS  => 1
	     );

# Lookup table for the different fields in the match
my %LUT = (
	   score => 0,

	   pSub => 1,
	   pDel => 2,
	   pIns => 3,

	   aName   => 4,
	   aStart  => 5,
	   aEnd => 6,
	   aRemain => 7,

	   strand  => 8,

	   bName   => 9,
	   bStart  => 10,
	   bEnd => 11,
	   bRemain => 12
	   );


sub new {
    my $pkg = shift;

    return bless {
	aFile => $_[0],
	bFile => $_[1],
	minM => $_[2],
	minS => $_[3],
	hit    => [],
	active => []
	}, $pkg;
}

# 0    1    2    3    4        5     6     7       8 9        10      11    12

# 98   0.00 0.00 0.00 130N4    1     104   84065   W 92M18    68800   68903 0
# 98   0.00 0.00 0.00 92M18    68800 68903 0       W 130N4    1       104   84065
# 8251 0.00 0.00 0.00 130N4    1     84169 0       W 130N4    1       84169 0
# 103  0.00 0.00 0.00 CFAT5    20771 20874 (0)     W A1280    1       104   (22149)
# 103  0.00 0.00 0.00 CFAT5    20771 20874 (0)     C A1280.RC (0)     22253 22150
# 103  0.00 0.00 0.00 CFAT5.RC 1     104   (20770) C A1280    (22149) 104   1

sub hit {
    my $match = shift;

    if (@_ == 13) {
	push( @{$match->{'hit'}}, [@_]);
	push( @{$match->{'active'}}, $#{$match->{'hit'}} );
    } else {
	croak qq(Bad number of elements in "@_");
    }
}

# Give list of active indices, or list of indices
# for all hits if no filter has been applied
sub list {
    my $match = shift;
    return @{$match->{'active'}};
}

# Restore active indices to list all hits
sub unfilter {
    my $match = shift;
    $match->{'active'} = [ (0..$#{$match->{'hit'}}) ];
    return (0..$#{$match->{'hit'}});
}

# Takes a ref to a subroutine as its argument
# which returns true or false for the line
# given as an argument
sub filter {
    my $match = shift;
    my $sub = shift;
    my @rows = @{$match->{'active'}};
    my( @true );

    foreach my $row (@rows) {
	if ( &$sub($match, $row) ) {
	    push @true, $row;
	}
    }

    # Save the filter
    @{$match->{'active'}} = @true;
}

sub endHits {
    my $match = shift;
    my $row = shift;
    
    if ($match->aStart($row) == 1 or
	$match->aRemain($row) == 0 or
	$match->bStart($row) == 1 or
	$match->bRemain($row) == 0) {
	return 1;
    } else {
	return;
    }
}

# Return number of hits
sub count {
    my $match = shift;
    
    return scalar @{$match->{'hit'}};
}

my %funcNames = (
		 start  => ['aStart', 'bStart'],
		 end    => ['aEnd', 'bEnd'],
		 remain => ['aRemain', 'bRemain']
		 );

# Functions which provide access to data fields by
# sequence name, by wrapping AUTOLOAD
sub start {
    my $match = shift;
    _chooser( 'start', $match, @_ );
}
sub end {
    my $match = shift;
    _chooser( 'end', $match, @_ );
}
sub remain {
    my $match = shift;
    _chooser( 'remain', $match, @_ );
}
sub _chooser {
    my $call = shift;
    my $match = shift;
    my $name = shift;
    my( $a, $b, $i, @newActive, $func, %uniq );

    # Look through aName and bName columns for name
    foreach (@{$match->{'active'}}) {
	# Compare $name to aName field
	if ($name eq $match->{'hit'}[$_][4]) {
	    $a = 1; push( @newActive, $_ );
	}
	# Compare $name to bName field
	if ($name eq $match->{'hit'}[$_][9]) {
	    $b = 1;
	    # Add to list unless already have it
	    unless ($newActive[$#newActive] == $_) {
		push( @newActive, $_ );
	    }
	}
    }

    # Save the new active list
    $match->{'active'} = [ @newActive ];

    # Did we find the sequence name in the "a" or "b" column?
    if ($a) {
	if ($b) {
	    carp qq($name matches in both "a" and "b" columns -- choosing "a");
	}
	$i = 0;
    } elsif ($b) {
	$i = 1;
    }

    # Call function if found match to $name, or else return nothing
    if (defined $i) {
	$func = $funcNames{$call}[$i];
	$match->$func(@_);
    } else {
	#carp qq("$name" not found);
	return;
    }

}

sub AUTOLOAD {
    my $match = shift;
    my @rows = @_;
    my ($name) = $AUTOLOAD =~ /:([^:]+)$/;
    my( $field, @values );

    # Ignore DESTROY messages...
    return if $name eq 'DESTROY';

    unless (ref($match)) {
        croak "$match: not an object; can't use AUTOLOAD to generate method: $name ";
    }
    
    # Deal with requests for a parameter field
    if ($field = $param{$name}) {
	return $match->{$name};
    }
    
    # Resolve name of method
    $field = $LUT{$name}
    or croak qq(unknown method "$name");
    
    if (wantarray) {
	# Extract the requested values
	unless (@rows) {
	    # Get a list of all the row indices
	    @rows = @{$match->{'active'}};
	}
	# Make a vertical slice through the hits
	foreach my $row (@rows) {
	    push(@values, $match->{'hit'}[$row][$field]);
	}
	return @values;
    } else {
	# Return just one value in scalar context
	if (defined $rows[0]) {
	    # Get field from row requested
	    return $match->{'hit'}[$rows[0]][$field];
	} else {
	    # Just get value from first row in active list
	    my $row = $match->{'active'}[0];
	    return $match->{'hit'}[$row][$field];
	}
    }
}

1;


__END__

=head1 NAME - CrossMatch.pm

=head1 DESCRIPTION

Module to provide object-oriented access to Phil Green's B<cross_match>
Smith-Waterman alignment program.

=head1 SYNOPSIS

	use CrossMatch;
	
	# Create a factory object
	$matcher = CrossMatch::Factory->new( '/nfs/disk2001/this_dir',
	                                     '/home/jgrg/that_dir' );
	$matcher->minMatch( 80 );

	# Match two fasta fomat sequence files, generating
        # a match object
	$cm = $matcher->crossMatch( 'dJ334P19.seq', 'cB49C12.aa' );

=head1 FUNCTIONS

=over 4

=item new

Create a new CrossMatch factory object.  Optional arguments to new is a
list of directories to search (see B<dir> below).

=item crossMatch

Do a cross_match on the two B<fasta formatted> sequence files supplied
as agruments, returning a B<CrossMatch> object.  The B<CrossMatch>
object records the parameters used in the match, and the match data
returned my cross_match (if any).

I<Note:> The two fasta files may each contain multiple sequences.

=item dir

Change the list of directories in which B<crossMatch> searches for
files.  The list of directories supplied to B<dir> completely replaces
the existing list.  Directories are searched in the order supplied.

This list defaults to the current working directory when the
B<CrossMatch::Factory> object is created.

=item extn

A convenience function which allows you to supply a list of filename
extensions to be considered by B<crossMatch> when finding a file to
pass to cross_match.  For example:

	$matcher->extn( 'seq', 'aa' );
	$cm = $matcher->crossMatch( 'dJ334P19', 'cB49C12' );

B<crossMatch> will look for files "dJ334P19.seq", "dJ334P19.aa".  If
no B<extn>s had been set, then only "dJ334P19" would have been
searched for.

=item minMatch

Set the B<minmatch> parameter passed to cross_match.  Defaults to 50.

=item minScore

Set the B<minscore> parameter passed to cross_match.  Defaults to 50.

=back

=head1 CrossMatch

Data stored in B<CrossMatch> objects, generated by a
B<CrossMatch::Factory>, can be retrieved with a variety of queries.
All the matches are stored internally in an array, in the order in
which they were generated by cross_match.  You can apply filters and
sorts (sorts are not yet implemented) to the object, which re-order or
hide matches.  These operations save their changes by altering the
active list, which is simply an array of indices which refer to
elements in the array of matches.  The data access funtions all
operate through this active list, and the active list can be reset to
show all the matches found by calling the B<unfilter> method.

=over 4

=item SYNOPSIS

	my $numberOfHits = $cm->count;
	foreach my $hit ($cm->list) {
	    print $cm->score($hit);
        }

	$firstName = $cm->aName;
	@allscores = $cm->score();
	@someSores = $cm->score(0,3,4);


=item FUNCTIONS

=over 4

=item count

Returns a count of the number of hits in the active list

=item list

Returns a list of array indices for the current list.

=item filter

	$cm->filter(\&Match::endHits);

Takes a reference to a subroutine (or an anonymous subroutine), and
alters the active list to contain only those hits for which the
subroutine returns true.  The subroutine should expect to be passed a
B<CrossMatch> object, and an integer corresponding to the index of a
match.  Applying a series of filters to the same object removes
successively more objects from the active list.

=item unfilter

Resets the active list to include all the hits, in the order in which
they were generated by cross_match.  Returns a list of the active
indices.

=item FILTERS

=over 4

=item endHits

A filter which returns true if one of the sequnces in a hit has its
end in the hit.

=back

=back

=item PARAMETERS

The parameters used by cross_match when doing the match can be
retrieved from the resulting Match object.  The two sequences matched
are labelled I<a> and I<b>, where I<a> is the sequence from the first
file passed to B<crossMatch>, and I<b>, the sequence from the second
file.  For example:

	$path_to_a = $cm->aFile;

retrieves the path of the file supplied as the first argument to
cross_match.

=over 4

=item aFile bFile

The full paths to the files used for sequences I<a> and I<b>.

=item minM

The minmatch parameter used.

=item minS

The minscore parameter used.

=back

=item DATA FIELDS

Syntax:

	$field = $match->FIELD();
	$field = $match->FIELD(INDEX);
	@fields = $match->FIELD();
	@fields = $match->FIELD(LIST);

Examples:

	$firstScore = $match->score(0);
	@aEnds = $match->aEnd();

These methods provide access to all the data fields for each hit found
by cross_match.  Each of the fields is described below.  In scalar
context, a single data field is returned, which is the field from the
first row in the active list if no I<INDEX> is given.  In array
context it returns a list with the I<FIELD> from all the hits in the
active list when called without arguments, or from the hits specified
by a I<LIST> of indices.

=over 4

=item score

The Smith-Waterman score of a hit, adjusted for the complexity of the
matching sequence.

=item pSub

The percent substitutions in the hit.

=item pDel

The percent deletions in the hit (sequence I<a> relative to sequence I<b>).

=item pIns

The percent insertions in the hit (sequence I<a> relative to sequence I<b>).

=item aName bName

ID of the first sequence (sequence I<a>) and second seqeunce (I<b>) in
the match, respecitvely.

=item aStart bStart

Start of hit in I<a>, or I<b>, respectively.

=item aEnd bEnd

End of hit in I<a>, or I<b>, respectively.

=item aRemain

Number of bases in I<a> after the end of the hit.  ("0" if the hit
extends to the end of I<a>.)

=item bRemain

The equivalent for sequence I<b> of aRemain, but note that if the
strand of I<b> which matches is the reverse strand, then this is the
number of bases left in I<b> after the hit, back towards the beginning
of I<b>.

=item strand

The strand on sequence I<b> which was matched.  This is "B<W>" for the
forward (or parallel or B<Watson>) strand, and "B<C>" for the reverse
(or anti-parallel or B<Crick>) strand.

=back

=item QUERYING BY SEQUENCE NAME

These methods allow access to the B<start>, B<end>, and B<remain>
fields for those occasions where you know the name of your sequence,
but don't necessarily know if it is sequence I<a> or I<b>.  Syntax and
behaviour is the same as for the I<DATA FIELDS> functions, but the
first argument to the function is the name of the sequence you want to
get data about.

I<Note:> These methods perform a filtering operation, reducing the
active list to only those hits which match the name given (in either
the I<a> or I<b> columns).  You'll therefore need to B<unfilter> your
match if you need data from hits which have now become hidden.

If the sequence name is found in both columns I<a> and I<b>, then a
warning is printed, and the I<a> column is chosen.

For example, suppose you have matched two fasta files containing only
the sequences cN75H12 and bK1109B5 (in that order), then the following
calls retrieve the same data:

	@ends = $match->aEnd();
	@ends = $match->end('cN75H12');

	$start = $match->bStart(0);
	$start = $match->start('bK1109B5', 0);

A warning is printed to STDERR if a match contains hits from the name
supplied in both columns, and only hits from the a column are
returned.

=over 4

=item start end remain

Access the aStart or bStart, aEnd or bEnd, and aRemain or bRemain
fields, depending upon wether the name supplied matches the aName or
bName fields respectively.

=back

=head1 BUGS


=head1 AUTHOR

B<James Gilbert> email jgrg@sanger.ac.uk
