
#
# Perl Module for SPEM::Entry.pm
#
# Cared for by Matthew Pocock <mrp@sanger.ac.uk>
#
# $Header: /nfs/disk45/mrp/publicEMBL/Entry.pm,v 1.6 1998/11/24 15:28:18 mrp Exp $
#
#Copyright Genome Research Limited (1997). Please see information on licensing in LICENSE

package SPEM::Entry;

use vars qw($AUTOLOAD @ISA @EXPORT_OK %fields $verbose);
use Exporter;
use Carp;
use strict;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = qw();

# @ISA has our inheritance
@ISA = ( 'Exporter' );
    
my %fields = (
	      #Insert field names here as field => undef,
	      lineEntries => undef
	      );

SPEM::Entry->verboseErrors(1);

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    my $self = { '_permitted' => \%fields,
		 %fields };

    bless $self, $class;
   
    $self->lineEntries([]);   # put an array into lineEntries to stoor the line groups

    return $self;
}

#
# Autoload up my field name set/get
#
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) || carp "Ok, $self is not an object of mine!";
    my $name = $AUTOLOAD;

    # don't propagate DESTROY messages...

    $name =~ /::DESTROY/ && return;

    $name =~ s/.*://; #get only the bit we want
    unless (exists $self->{'_permitted'}->{$name} ) {
	carp "In type $type, can't access $name - probably passed a wrong variable into Sequence  ";
    }
    if (@_) {
	return $self->{$name} = shift;
    } else {
	return $self->{$name}
    }
}

sub addLineGroup {
    my $self = shift;
    my $what = shift;
    my $where = shift;
    my $lines = $self->lineEntries;

    unless(defined $what) {
	carp 'Use: $entry->($what [, $where])';
	return undef;
    }

    (not defined $where) && return push @$lines, $what;
    $where =~ m/end/i && return push @$lines, $what;
    $where =~ m/start/i && return unshift @$lines, $what;
    $where =~ m/a-z/i && do {
	carp "where parameter must be START, STOP, or an integer";
	return undef;
    };
    ($where < @$lines) && (return splice @$lines, $where, 0, $what)
	|| return push @$lines, $what;

}

sub removeLineGroup {
    my $self = shift;
    my $where = shift;

    unless(defined $where) {
	carp 'Use: $entry->($where)';
	return undef;
    }

    return splice @{$self->lineEntries}, $where, 1;
}

sub get {
    my $self = shift;
    my $where = shift || 0;

    # index within this object
    ($where < @{$self->lineEntries} && $where >= 0) &&
	return $self->lineEntries->[$where];

    # index outside - return undef after warning
  SPEM::Entry->verboseErrors() &&
      carp "You have tried to index an item that does not exist in this entry.";
    return undef;
}

sub _indexFromStart {
    my $self = shift;
    my $lineType = shift;
    my $index = shift || 0;
    my $lines = $self->lineEntries();

    while(ref $lines->[$index] ne $lineType) {
	$index++;
	if($index > @$lines) {
	  SPEM::Entry->verboseErrors() &&
	      carp "could not find any $lineType lines.";
	    return undef;
	}
    }

    return $index;
}

sub _indexFromEnd {
    my $self = shift;
    my $lineType = shift;
    my $lines = $self->lineEntries();
    my $index = shift || @$lines;

    while(ref $lines->[$index] ne $lineType) {
	$index--;
	if($index == -1) {
	  SPEM::Entry->verboseErrors() &&
	      carp "could not find any $lineType lines.";
	    return undef;
	}
    }
    
    return $index;
}

sub getLine {
    my $self = shift;
    my $lineType = shift;
    my $index = shift;

    $index = $self->_indexFromStart($lineType, $index);
    return (defined $index)
      ? $self->get($index)
	: undef;
}

sub all {
  my $self = shift;
  my $lineType = shift;
  my $index = 0;
  my $line;
  my @lines = ();
  my $error = $self->verboseErrors();
  $self->verboseErrors(undef);

  while(1) {
    defined($index = $self->_indexFromStart($lineType, $index)) || last;
    defined($line=$self->get($index)) || last;
    push @lines, $line;
    $index++;
  }

  $self->verboseErrors($error);
  return @lines;
}
  
sub toStream {
    my $self = shift;
    my $fh = shift;
    my $line;

    foreach $line (@{$self->lineEntries()}) {
	$self->writeLine($fh, $line);
    }
    print $fh "//\n";
}
	
sub ID {
    my $self = shift;
    confess "You have tried to call the pure virtual function ID for object $self.\nYou should override this method in all subclasses to return the unique ID of the entry";
}

sub writeLine {
    my $self = shift;
    my $fh = shift;
    my $line = shift;
    confess "You have tried to call the pure virtual function writeLine for the object $self.\nYou should override this method in all subclasses to call the apropreate toStream method for a line.";
}

sub verboseErrors {
    my $thing = shift;
    if(@_) {
	return $verbose = shift;
    } else {
	return $verbose;
    }
}

1;  # says use was ok
__END__

=head1 NAME

SPEM::Entry.pm

=head1 DESCRIPTION

B<SPEM::Entry> is part of the core modules that deal with parsing and
representing SWISSPROT and EMBL formatted information. See <SPEM::Spem>
for an overview. Throughout this documentation, the concrete example
of L<EMBL::Entry> will be used for illustration.

B<SPEM::Entry> is a base class for objects which represent a single
flatfile entry and provides an api for their modification, searching,
sorting etc.

You can largely 'inherit' this documentation into a derived class. However,
use your descretion, and change any package names that seem apropreate to
ammend.

=head1 AUTHOR

B<Matthew Pocock> Email mrp@sanger.ac.uk

=head1 Member variables

=over

=item lineEntries

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

The array of line groups that make up this entry. Edit this and the entry will change.
Each line group is a single logical group of lines (the ID line, or all of the feature
table lines, for example). Functions are provided to access this list in a more friendly
manner, so for most purposes you should not need to worry about this.

=back

=head1 Functions

=head2 Line Editing

=over

=item addLineGroup

=for htmlAdd a line group object to this entry.

Optionaly specify where to add it.

    $entry->addLineEntry($what [, $where]);

If where is absent, then the line group will be added at the end of the entry.
If it is START or END, it will be added accordingly.  If it is a number, then
this is taken to be an index (starting at 0) for where to add it.  Should the
index be outside the range of the entry, it will be added at the end.

=item removeLineGroup

Removes the line group at the specified index. Returns the object removed,
or C<undef>.

    $entry->removeLineGroup($where);

=back

=head2 Retrieve lines

=over

=item get

Returns the line group at the specified position.

    $lg = $entry->get($where);

The C<where> index starts at zero.  If you try to index a group beyond the end
of this object then a clear message to that effect will be raised, and the
return value will be  C<undef>.

=item getLine

Returns a line group object of a given type. Use:

    $line = $entry->getLine('EMBL::CC');

to get the first embl comment line. Optionaly, a seccond parameter can be passed which
causes the search for the next line to begin at that index.

    $line = $entry->getLine('EMBL::CC', 8);

will return the first embl comment line on or after the line with index 8.

=back

=head2 Private Scanning Functions

=over

=item _indexFromStart

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

This function takes a string that represents a class type, and returns an
index that can be used with C<get> which represents the first line of
that type in this object. Optionaly, you can parse it a starting index.  The
search will then start at that position (as the perl built in function C<index>
does).

=item _indexFromEnd

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

This function takes a string that represents a class type, and returns an
index that can be used with C<get> which represents the last line of
that type in this object. Optionaly, you can parse it a starting index.  The
search will then start at that position (as the perl built in function C<index>
does).

=item verboseErrors

=for html
<img src = "http://www.sanger.ac.uk/~mrp/magic.gif"><p>

  SPEM::Entry->verboseErrors(1);         # turns verbose errors on (default)
  SPEM::Entry->verboseErrors(undef);     # turns verbose errors off
  $verbose = SPEM::Entry->verboseErrors; # returns the verbosity

=back

=head2 Pure Virtual Functions

These are the functions that make it possible for the algorithms for reading,
writing and stooring of the SPEM objects to be encapsulated. They must be
overridden, or your program will not run - printing a long error message.

=over

=item ID

You must provide an ID function when subclassing SPEM::Entry.
This should return a string or number that represents the unique ID of the entry in the database.
This will be used by SPEM::Set to stoor and retrieve the object.

In <EMBL::Entry#ID>, a scan is made using C<_indexFromStart> for lines of type
L<EMBL::ID>. This line is then interogated with L<EMBL::ID#entryname> to find
the actual ID. You will probably wish to use this pattern.

=item writeLine

You must provide a writeLine function. This is called by C<toStream> for each line group.
It will be called as $self->writeLine($fh, $lineGroup).  It is up to you to call the
apropriate function on $lineGroup, but I would invisige you using it like this (again an EMBL
example).

=for html
<PRE>
sub writeLine {
    $self = shift;
    $fh = shift;
    $lineGroup = shift;
 
    $lineGroup->toEMBLStream($fh);
}
=for html
</PRE>

Where in this example the lineGroup objects will all have a toEMBLStream function that
writes them out in EMBL format.

It follows from this that some ubiquitous object (such as L<Sequence>) may have a toEMBLStream,
a toSWISSStream etc., one for each database you have a module set for. This allows you to decide
in the Entry module which function will actually be used.

=back
