
#
# Perl Module for TempFiles
#
# Cared for by Simon Dear <sd@sanger.ac.uk>
# Ported to 'official sanger' by Ewan
#
package TempFiles;

use vars qw($AUTOLOAD @ISA);
use Exporter;
use Carp;
use strict;

#
# Place functions/variables you want to *export*, ie be visible from the caller package into @EXPORT_OK
#

@EXPORT_OK = ();

#
# @ISA has our inheritance.
#

@ISA = ( 'Exporter' );

package TempFiles;

##########################################
#
# Construct...
#
##########################################

sub new { 

    my $type=shift;

    my $self={};
    $type->setDefaults($self);

    my %args = @_;
    foreach (keys%args) {
        $self->{$_} = $args{$_};
    }

    @{$self->{'temp_files'}} = ();

    bless $self, $type;

}

sub setDefaults {

    my ($type, $param_ref)=@_;

    $param_ref->{'dir'} = '/tmp/';         # Directory to write files
    $param_ref->{'prefix'} = '_TempFiles'; # All temp files will carry this prefix
    $param_ref->{'suffix'} = "_$$";        # Process id suffix for uniqueness
    $param_ref->{'file_count'} = 0;        # Allocation count
}


##########################################
#
# Methods
#
##########################################


############################################################
# $temp_file_name=&getTemp;
#
# Allocate and register temp file name.
#
# Facilitates for temp file deletion on error.
#
############################################################

sub getTemp {

    my $type = shift;
    my %args = @_;

    my $dir = $args{'-dir'} || $type->{'dir'};
    my $prefix = $args{'-prefix'} || $type->{'prefix'};
    my $suffix = $args{'-suffix'} || $type->{'suffix'};
    my $count = $type->{'file_count'};

    my $new_temp_filename="$dir$prefix$count$suffix";

    $type->{'file_count'}++;
    push(@{$type->{'temp_files'}}, $new_temp_filename);

    $new_temp_filename;
}



############################################################
# destroyTemp($file)
#
# Deletes temporary file(s).
#
# If no args, *all* temp files are removed
#
############################################################

sub destroyTemp {

    my $type = shift;
    my(@to_delete)=@_;

    if ($#to_delete==-1) {
        if (defined(@{$type->{'temp_files'}})) {
            # Delete all temporary files listed in @{$type->{'temp_files'}}
            foreach (@{$type->{'temp_files'}}) {
                if (-e $_) { # Not all allocated files get wriiten to.
                    unlink("$_") || carp "\nWARNING: Unable to remove temporary file: \"$_\"$!\n";
                }
            }
            undef(@{$type->{'temp_files'}});
        }
    } else {
	my $file;
        foreach $file (@to_delete) {
	    my $i;
            if (-e "$file") {
                unlink("$file") || carp "\nWARNING: Unable to remove temporary file: \"$_\"$!\n";
            }
            for($i=0;$i<=$#{$type->{'temp_files'}};$i++) {
                if (${$type->{'temp_files'}}[$i] eq $file) {
                    splice(@{$type->{'temp_files'}},$i,1);
                    $i--;
                }
            }
        }
    }
}



1;  # says use was ok
__END__

=head1 NAME

TempFiles

=head1 DESCRIPTION

A module for handling Temporary files.

=head1 AUTHOR

B<Simon Dear> Email sd@sanger.ac.uk

=head1 Example

	#
	# Typical construction:
	# 
    	
	my $tfiles=TempFiles->new();  # Initialise package use default directory. prefix, etc..

	my $cafile = $tfiles->getTemp(-prefix => 'caf');  #$cafile gets set to something like /tmp/caf0_17348


	#
	# END blocks are guarented to run on exits
	#

	END { $tfiles->destroyTemp if defined $tfiles; }   #Trash all temp files upon any exit.

=head1 Methods

=over

=item new

Makes a new TempFiles object. Can take 'tag' => 'value' arguments of 'dir' 'prefix' 'suffix'. 
These get set to defaults for off.

=item getTemp

gets a string to a temporary file with suitable attributes. Use the -prefix => 'xxx' seems
the common idiom here

=item destroyTemp 

destroys Temp files. If you give an argument, it will destroy that file only. No args
will destroy all Temporary files

=back








